#
# Copyright 2013-2016 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import re
import sys
import unittest
import tests.unit.support
import ufw.parser

class ParserTestCase(unittest.TestCase):
    def setUp(self):
        self.parser = ufw.parser.UFWParser()

        # Basic commands
        for i in ['enable', 'disable', 'help', '--help', 'version', \
                  '--version', 'reload', 'reset' ]:
            self.parser.register_command(ufw.parser.UFWCommandBasic(i))

        # Application commands
        for i in ['list', 'info', 'default', 'update']:
            self.parser.register_command(ufw.parser.UFWCommandApp(i))

        # Logging commands
        for i in ['on', 'off', 'low', 'medium', 'high', 'full']:
            self.parser.register_command(ufw.parser.UFWCommandLogging(i))

        # Default commands
        for i in ['allow', 'deny', 'reject']:
            self.parser.register_command(ufw.parser.UFWCommandDefault(i))

        # Status commands ('status', 'status verbose', 'status numbered')
        for i in [None, 'verbose', 'numbered']:
            self.parser.register_command(ufw.parser.UFWCommandStatus(i))

        # Show commands
        for i in ['raw', 'before-rules', 'user-rules', 'after-rules', \
                  'logging-rules', 'builtins', 'listening', 'added']:
            self.parser.register_command(ufw.parser.UFWCommandShow(i))

        # Rule commands
        rule_commands = ['allow', 'limit', 'deny' , 'reject', 'insert', \
                         'delete']
        for i in rule_commands:
            self.parser.register_command(ufw.parser.UFWCommandRule(i))
            self.parser.register_command(ufw.parser.UFWCommandRouteRule(i))

    def tearDown(self):
        pass

    def test_ufwcommand_parse_empty(self):
        '''Test UFWCommand.parse([])'''
        c = ufw.parser.UFWCommand('basic', 'status')
        tests.unit.support.check_for_exception(self, ValueError, \
                                                   c.parse,
                                                   [])

    def test_ufwcommand_help(self):
        '''Test UFWCommand.help()'''
        c = ufw.parser.UFWCommand('basic', 'status')
        tests.unit.support.check_for_exception(self, ufw.common.UFWError, \
                                                   c.help,
                                                   [])

    def test_ufwcommand_parse(self):
        '''Test UFWCommand.parse()'''
        c = ufw.parser.UFWCommand('basic', 'status')
        pr = c.parse(['status'])
        self.assertEquals('status', pr.action, "%s != 'status'" % (pr.action))

    def test_ufwcommandbasic_parse_with_arg(self):
        '''Test UFWCommand.parse() - basic with arg'''
        parser = ufw.parser.UFWParser()
        c = ufw.parser.UFWCommandBasic('enable')
        tests.unit.support.check_for_exception(self, ValueError, \
                                                   c.parse,
                                                   ['enable', 'OpenSSH'])

    def test_ufwparser_response(self):
        '''Test UFWParserResponse.str()'''
        cmd = 'rule allow 22'
        pr = self.parser.parse_command(cmd.split())
        s = str(pr)
        search = repr("action='allow',iptype='both'," + \
                      "rule='-p all --dport 22 -j ACCEPT',type='rule'\n")
        self.assertTrue(s == search, "'%s' != '%s'" % (s, search))
        self.assertFalse(pr.dryrun)
        self.assertFalse(pr.force)

        cmd = '--dry-run rule allow 22'
        pr = self.parser.parse_command(cmd.split())
        s = str(pr)
        search = repr("action='allow',iptype='both'," + \
                      "rule='-p all --dport 22 -j ACCEPT',type='rule'\n")
        self.assertTrue(s == search, "'%s' != '%s'" % (s, search))
        self.assertTrue(pr.dryrun)
        self.assertFalse(pr.force)

        cmd = '--force rule allow 22'
        pr = self.parser.parse_command(cmd.split())
        s = str(pr)
        search = repr("action='allow',iptype='both'," + \
                      "rule='-p all --dport 22 -j ACCEPT',type='rule'\n")
        self.assertTrue(s == search, "'%s' != '%s'" % (s, search))
        self.assertTrue(pr.force)

    def test_ufwparser_register_command(self):
        '''Test UFWParser.register_command()'''
        parser = ufw.parser.UFWParser()
        c = ufw.parser.UFWCommandBasic('enable')
        parser.register_command(c)
        self.assertTrue('basic' in parser.commands)
        self.assertTrue('enable' in parser.commands['basic'])

        # Register an already existing command
        tests.unit.support.check_for_exception(self, ufw.common.UFWError, \
                                                   parser.register_command,
                                                   c)

    def test_ufwparser_register_command_none(self):
        '''Test UFWParser.register_command()'''
        parser = ufw.parser.UFWParser()
        c = ufw.parser.UFWCommandBasic('enable')
        c.command = None
        parser.register_command(c)
        self.assertTrue('basic' in parser.commands)
        self.assertTrue('basic' in parser.commands['basic'])

    def test_ufwparser_allowed_command(self):
        '''Test UFWParser.allowed_command()'''
        # Valid commands
        cmd = 'enable'
        res = self.parser.allowed_command('basic', cmd)
        search = cmd.lower()
        self.assertTrue(res == search, "'%s' != '%s'" % (res, search))

        cmd = 'enable'
        res = self.parser.allowed_command('basic', cmd.upper())
        search = cmd.lower()
        self.assertTrue(res == search, "'%s' != '%s'" % (res, search))

        # Invalid commands
        tests.unit.support.check_for_exception(self, ValueError, \
                                                   self.parser.allowed_command,
                                                   'basic', 'nonexistent')

        tests.unit.support.check_for_exception(self, ValueError, \
                                                   self.parser.allowed_command,
                                                   'nonexistent', 'allow')

    def test_ufwcommand_rule_get_command(self):
        '''Test UFWCommand(Route)Rule.get_command()'''
        count = 0
        cmds = tests.unit.support.get_sample_rule_commands_simple()
        cmds += tests.unit.support.get_sample_rule_commands_extended()
        cmds += tests.unit.support.get_sample_rule_commands_extended(v6=True)
        cmds += [
                 ['rule', 'reject', 'from', 'any', 'app', 'Apache'],
                 ['rule', 'reject', 'from', 'any', 'port', 'smtp'],
                 ['route', 'reject', 'from', 'any', 'app', 'Apache'],
                 ['route', 'reject', 'from', 'any', 'port', 'smtp'],
                 ['route', 'allow', 'out', 'on', 'eth1', 'in', 'on', 'eth0'],
                ]
        errors = []

        for cmd in cmds:
            count += 1
            #print(" ".join(cmd))
            # Note, parser.parse_command() modifies its arg, so pass a copy of
            # the cmd, not a reference
            pr = self.parser.parse_command(cmd + [])
            if cmd[0] == 'rule':
                res = ufw.parser.UFWCommandRule.get_command(pr.data['rule'])
            else:
                res = ufw.parser.UFWCommandRouteRule.get_command(
                        pr.data['rule'])

            # First, feed the res rule into parse() (we need to split the
            # string but preserve quoted substrings
            if sys.version_info[0] < 3:
                test_cmd = [cmd[0]] + \
                           [p.strip("'").encode('utf-8') for p in re.split("( |'.*?')",
                                                           res) if p.strip()]
            else:
                test_cmd = [cmd[0]] + \
                           [p.strip("'") for p in re.split("( |'.*?')",
                                                           res) if p.strip()]
            try:
                self.parser.parse_command(test_cmd + [])
            except ufw.common.UFWError:
                self.assertTrue(False,
                                "get_command() returned invalid rule:\n" + \
                                " orig=%s\n pr.data['rule']=%s\n result=%s" % \
                                (cmd, pr.data['rule'], test_cmd))

            # Next, verify the output is what we expect. We need to massage the
            # cmd_compare output a bit first since many rules can be expressed
            # using the same syntax. Eg, these are all the same rule and
            # get_command() typically outputs the simplest form:
            #  ufw allow 22
            #  ufw allow in 22
            #  ufw allow to any port 22
            #  ufw allow from any to any port 22
            #  ufw rule allow 22
            #  ufw rule allow in 22
            #  ufw rule allow to any port 22
            #  ufw rule allow from any to any port 22

            # Note, cmd_compare contains the rules we get from
            # tests.unit.support.get_sample_rule_commands*
            cmd_compare = []
            comment = ""  # store off command so we can add it at the end after
                          # the massaging
            if 'comment' in cmd:
                comment_idx = cmd.index('comment')
                comment = cmd[comment_idx + 1]
                del cmd[comment_idx + 1]
                del cmd[comment_idx]

            for i in cmd:
                if ' ' in i:  # quote anything with a space for comparisons
                    cmd_compare.append("'%s'" % i)
                else:
                    cmd_compare.append(i)

            # remove 'in' on rules without an interface
            if 'in' in cmd_compare and 'on' not in cmd_compare:
                cmd_compare.remove('in')

            # use '1/tcp' instead of 'tcpmux' for simple rules and
            # 'port 1 proto tcp' for extended
            if 'tcpmux' in cmd_compare:
                if 'to' in cmd_compare or 'from' in cmd_compare:  # extended
                    cmd_compare[cmd_compare.index('tcpmux')] = '1'
                    if 'proto' not in cmd_compare:
                        cmd_compare.append('proto')
                        cmd_compare.append('tcp')
                    if 'tcpmux' in cmd_compare:  # can have 2 in extended rules
                        cmd_compare[cmd_compare.index('tcpmux')] = '1'
                else:  # simple
                    cmd_compare[cmd_compare.index('tcpmux')] = '1/tcp'

            # use '21/udp' instead of 'fsp' for simple rules and
            # 'port 21 proto udp' for extended
            if 'fsp' in cmd_compare:
                if 'to' in cmd_compare or 'from' in cmd_compare:  # extended
                    cmd_compare[cmd_compare.index('fsp')] = '21'
                    if 'proto' not in cmd_compare:
                        cmd_compare.append('proto')
                        cmd_compare.append('udp')
                    if 'fsp' in cmd_compare:  # can have 2 in extended rules
                        cmd_compare[cmd_compare.index('fsp')] = '21'
                else:  # simple rule
                    cmd_compare[cmd_compare.index('fsp')] = '21/udp'

            # use 'port 25 proto tcp' in extended rules
            if 'smtp' in cmd_compare and 'proto' not in cmd_compare:
                cmd_compare[cmd_compare.index('smtp')] = '25'
                cmd_compare.append('proto')
                cmd_compare.append('tcp')

            # remove 'from any' clause when used without port or app
            if 'from' in cmd_compare and \
               cmd_compare[cmd_compare.index('from') + 1] == 'any' and \
               (len(cmd_compare) - 2 == cmd_compare.index('from') or \
                (cmd_compare.index('from') + 2 < len(cmd_compare) and \
                 cmd_compare[cmd_compare.index('from') + 2] != 'port' and \
                 cmd_compare[cmd_compare.index('from') + 2] != 'app')):
                del cmd_compare[cmd_compare.index('from') + 1]
                cmd_compare.remove('from')

            # remove 'to any' clause when used without port or app when 'from'
            # 'proto' or 'on' is present ('from' will not be 'any' because of
            # above)
            if ('from' in cmd_compare or 'proto' in cmd_compare or \
                'on' in cmd_compare) and 'to' in cmd_compare and \
               cmd_compare[cmd_compare.index('to') + 1] == 'any' and \
               (len(cmd_compare) - 2 == cmd_compare.index('to') or \
                (cmd_compare.index('to') + 2 < len(cmd_compare) and \
                 cmd_compare[cmd_compare.index('to') + 2] != 'port' and \
                 cmd_compare[cmd_compare.index('to') + 2] != 'app')):
                del cmd_compare[cmd_compare.index('to') + 1]
                cmd_compare.remove('to')

            # remove 'to any' if no 'from' clause (ie, convert extended to
            # simple)
            if 'to' in cmd_compare and 'from' not in cmd_compare and \
               cmd_compare[cmd_compare.index('to') + 1] == 'any' and \
               cmd_compare.index('to') + 2 < len(cmd_compare) and \
               'on' not in cmd_compare:
                if 'port' in cmd_compare:
                    port = "%s" % cmd_compare[cmd_compare.index('port') + 1]
                    if 'proto' in cmd_compare:
                        port += "/%s" % \
                                cmd_compare[cmd_compare.index('proto') + 1]
                    del cmd_compare[cmd_compare.index('proto') + 1]
                    cmd_compare.remove('proto')
                    del cmd_compare[cmd_compare.index('port') + 1]
                    cmd_compare.remove('port')
                    del cmd_compare[cmd_compare.index('to') + 1]
                    cmd_compare.remove('to')
                    cmd_compare.append(port)
                elif 'app' in cmd_compare:
                    del cmd_compare[cmd_compare.index('to') + 2]
                    del cmd_compare[cmd_compare.index('to') + 1]
                    cmd_compare.remove('to')

            # add back 'to any' if have no 'to', 'from' or 'on' and have either
            # proto or the last entry in cmd_compare indicates generic extended
            # rule
            generics = ['in', 'out', 'allow', 'deny', 'reject', 'limit']
            if 'to' not in cmd_compare and 'from' not in cmd_compare and \
               'on' not in cmd_compare and ('proto' in cmd_compare or \
               cmd_compare[-1].startswith('log') or \
               cmd_compare[-1] in generics):
                if 'proto' in cmd_compare:
                    cmd_compare.insert(cmd_compare.index('proto'), "to")
                    cmd_compare.insert(cmd_compare.index('proto'), "any")
                else:
                    cmd_compare.append("to")
                    cmd_compare.append("any")

            # flip 'in on' and 'out on' for route rules ('in on' is always
            # listed first
            if cmd_compare[0] == 'route' and \
               'out' in cmd_compare and 'in' in cmd_compare and \
               cmd_compare.index('out') < cmd_compare.index('in'):
                   tmp_out_idx = cmd_compare.index('out')
                   tmp_outif = cmd_compare[tmp_out_idx + 2]
                   tmp_in_idx = cmd_compare.index('in')
                   tmp_inif = cmd_compare[tmp_in_idx + 2]
                   cmd_compare[tmp_out_idx] = 'in'
                   cmd_compare[tmp_out_idx + 2] = tmp_inif
                   cmd_compare[tmp_in_idx] = 'out'
                   cmd_compare[tmp_in_idx + 2] = tmp_outif

            # add comment back
            if comment != "":
                cmd_compare.append('comment')
                compare_str = " ".join(cmd_compare)
                if sys.version_info[0] < 3:
                    compare_str += " '%s'" % comment.decode('utf-8')
                else:
                    compare_str += " '%s'" % comment
                cmd_compare.append(comment)
            else:
                compare_str = " ".join(cmd_compare)
            if "%s %s" % (cmd[0], res) != compare_str:
                errors.append(" \"%s %s\" != \"%s\" (orig=%s)" % (cmd[0], res,
                    compare_str, cmd))

            #print("Result: rule %s" % res)

        self.assertEquals(len(errors), 0,
                          "Rules did not match:\n%s\n(%d of %d)" % \
                          ("\n".join(errors), len(errors), count))
        print("%d rules checked" % count)

    def test_simple_parse(self):
        '''Test simple rule syntax'''
        count = 0
        cmds = tests.unit.support.get_sample_rule_commands_simple()
        for cmd in cmds:
            count += 1
            #print(" ".join(cmd))
            # Note, parser.parse_command() modifies its arg, so pass a copy of
            # the cmd, not a reference
            pr = self.parser.parse_command(cmd + [])

            # TODO: more tests here by sending the cmd and the pr to a helper
            action = cmd[1]
            self.assertEquals(action, pr.action, "%s != %s" % (action, \
                                                               pr.action))

            del_cmd = cmd + []
            del_cmd.insert(1, 'delete')
            #print(" ".join(del_cmd))
            # Note, parser.parse_command() modifies its arg, so pass a copy of
            # the del_cmd, not a reference
            pr = self.parser.parse_command(del_cmd + [])

            # TODO: more tests here by sending the cmd and the pr to a helper
            action = del_cmd[2]
            self.assertEquals(action, pr.action, "%s != %s" % (action, \
                                                               pr.action))
            ins_cmd = cmd + []
            ins_cmd.insert(1, 'insert')
            ins_cmd.insert(2, '1')
            #print(" ".join(ins_cmd))
            # Note, parser.parse_command() modifies its arg, so pass a copy of
            # the del_cmd, not a reference
            pr = self.parser.parse_command(ins_cmd + [])

            # TODO: more tests here by sending the cmd and the pr to a helper
            action = ins_cmd[3]
            self.assertEquals(action, pr.action, "%s != %s" % (action, \
                                                               pr.action))
        print("%d rules checked" % count)

    def test_misc_rules_parse(self):
        '''Test rule syntax - miscellaneous'''
        cmds = [
                ['rule', 'delete', '1'],
                ['rule', 'delete', 'allow', '22'],
                ['rule', 'deny', 'from', 'any', 'port', 'domain', 'to', \
                 'any', 'port', 'tftp'],
                ['rule', 'allow', 'to', 'any', 'proto', 'gre'],
                ['rule', 'deny', 'to', 'any', 'proto', 'ipv6'],
                ['rule', 'allow', 'to', 'any', 'proto', 'igmp'],
                ['rule', 'reject', 'to', 'any', 'proto', 'esp'],
                ['rule', 'deny', 'to', '224.0.0.1', 'proto', 'igmp'],
                ['rule', 'deny', 'in', 'on', 'eth0', 'to', '224.0.0.1', \
                 'proto', 'igmp'],
                ['rule', 'allow', 'in', 'on', 'eth0', 'to', '192.168.0.1', \
                 'proto', 'gre'],
                ['rule', 'deny', 'to', 'any', 'proto', 'ah'],
                ['rule', 'allow', 'out', 'on', 'br_lan'],
               ]
        count = 0
        for cmd in cmds:
            #print(" ".join(cmd))
            count += 1
            # Note, parser.parse_command() modifies its arg, so pass a copy of
            # the cmd, not a reference
            self.parser.parse_command(cmd + [])

    def test_rule_bad_syntax(self):
        '''Test rule syntax - bad'''
        cmds = [
                (['rule', 'insert', '1', 'allow'], ValueError),
                (['rule', 'insert', 'a', 'allow', '22'], ufw.common.UFWError),
                (['rule', 'insert', '0', 'allow', '22'], ufw.common.UFWError),
                (['rule', 'allow'], ValueError),
                (['rule'], ValueError),
                (['rule', 'allow', '22', 'in', 'on', 'eth0'],
                 ufw.common.UFWError),
                (['rule', 'allow', 'in', 'in', 'eth0', '22'],
                 ufw.common.UFWError),
                (['rule', 'allow', 'in', 'on', 'eth0', '22', 'log'],
                 ufw.common.UFWError),
                (['rule', 'allow', 'in', 'on', 'eth0', '22', 'log-all'],
                 ufw.common.UFWError),
                (['rule', 'allow', 'in', 'on', 'eth0', 'log', 'to', 'any', \
                  'port', '22', 'from', 'any', 'port', '123', 'proto', 'udp', \
                  'extra'], ValueError),
                (['rule', 'allow', '22/udp/p'], ufw.common.UFWError),
                (['rule', 'allow', '22:2e'], ufw.common.UFWError),
                (['rule', 'allow', '22/ipv6'], ufw.common.UFWError),
                (['rule', 'reject', 'in', 'on', 'eth0', 'port', '22'],
                 ufw.common.UFWError),
                (['rule', 'allow', 'to', 'any', '22'], ufw.common.UFWError),
                (['rule', 'allow', 'to', 'any', 'to', '22'],
                 ufw.common.UFWError),
                (['rule', 'allow', 'to', 'any', 'proto', 'nope'],
                 ufw.common.UFWError),
                (['rule', 'allow', 'in', 'on', '!eth0', 'to', 'any'],
                 ufw.common.UFWError),
                (['rule', 'allow', 'out', 'on', 'eth0:0', 'to', 'any'],
                 ufw.common.UFWError),
                (['rule', 'allow', 'in', 'on', '$eth', 'to', 'any'],
                 ufw.common.UFWError),
                (['rule', 'allow', 'in', 'eth0', 'to', 'any'],
                 ufw.common.UFWError),
                (['rule', 'allow', 'from', 'bad_address'],
                 ufw.common.UFWError),
                (['rule', 'allow', 'to', 'bad_address'], ufw.common.UFWError),
                (['rule', 'badcmd', 'to', 'any'], ValueError),
                (['rule', 'allow', 'port', '22'], ufw.common.UFWError),
                (['rule', 'allow', 'to', 'any', 'port', '22_23'],
                 ufw.common.UFWError),
                (['rule', 'allow', 'to', 'any', 'port', '22:_23'],
                 ufw.common.UFWError),
                (['rule', 'allow', 'to', 'any', 'port', '65536'],
                 ufw.common.UFWError),
                (['rule', 'allow', 'to', '::1', 'from', '127.0.0.1'],
                 ufw.common.UFWError),
                (['rule', 'allow', 'to', 'any', 'port', 'nonexistent'],
                 ufw.common.UFWError),
                (['rule', 'allow', 'from', 'any', 'port', 'nonexistent',
                  'proto', 'any'], ufw.common.UFWError),
                (['rule', 'allow', 'from', 'any', 'port', 'tftp', 'to', 'any',
                 'port', 'smtp'], ufw.common.UFWError),
                (['rule', 'deny', 'from', 'any', 'port', 'smtp', 'to', 'any',
                 'port', 'tftp', 'proto', 'any'], ufw.common.UFWError),
                (['rule', 'allow', 'nope', 'any', 'to', 'any'],
                 ufw.common.UFWError),
                (['rule', 'deny', 'to', 'any', 'port', 'tftp',
                  'proto', 'tcp'], ufw.common.UFWError),
                (['rule', 'deny', 'to', '::1', 'proto', 'ipv6'],
                 ufw.common.UFWError),
                (['rule', 'deny', 'to', '::1', 'proto', 'igmp'],
                 ufw.common.UFWError),
                (['rule', 'deny', 'to', 'any', 'port', '22', 'proto', 'ipv6'],
                 ufw.common.UFWError),
                (['rule', 'deny', 'to', 'any', 'port', '22', 'proto', 'igmp'],
                 ufw.common.UFWError),
                (['rule', 'deny', 'to', 'any', 'port', '22', 'proto', 'esp'],
                 ufw.common.UFWError),
                (['rule', 'deny', 'to', 'any', 'port', '22', 'proto', 'ah'],
                 ufw.common.UFWError),
                (['rule', 'deny', 'to', 'any', 'port', '22', 'proto', 'gre'],
                 ufw.common.UFWError),
                (['rule', 'allow', 'to', '192.168.0.0/16', 'app', 'Samba',
                  'from', '192.168.0.0/16', 'port', 'tcpmux'],
                  ufw.common.UFWError),
                (['route', 'badcmd', 'to', 'any'], ValueError),
                (['route', 'allow', 'in', '22'], ufw.common.UFWError),
                (['route', 'deny', 'out', '22'], ufw.common.UFWError),
                (['route', 'allow', 'to', '192.168.0.0/16', 'app', 'Samba',
                  'from', '192.168.0.0/16', 'port', 'tcpmux'],
                  ufw.common.UFWError),
               ]
        for cmd, exception in cmds:
            #print(" ".join(cmd))
            # Note, parser.parse_command() modifies its arg, so pass a copy of
            # the cmd, not a reference
            tests.unit.support.check_for_exception(self, exception,
                                                   self.parser.parse_command,
                                                   cmd + [])

    def test_extended_parse(self):
        '''Test extended rule syntax'''
        count = 0
        cmds = tests.unit.support.get_sample_rule_commands_extended()
        cmds6 = tests.unit.support.get_sample_rule_commands_extended(v6=True)
        for cmd in cmds + cmds6:
            count += 1
            #print(" ".join(cmd))
            # Note, parser.parse_command() modifies its arg, so pass a copy of
            # the cmd, not a reference
            pr = self.parser.parse_command(cmd + [])

            # TODO: more tests here by sending the cmd and the pr to a helper
            action = cmd[1]
            self.assertEquals(action, pr.action, "%s != %s" % (action, \
                                                               pr.action))

        print("%d rules checked" % count)

    def test_simple_bad_numeric_port(self):
        '''Test simple bad numeric port'''
        for port in ['-1', '1000000']:
            c = ['rule', 'allow', port]
            tests.unit.support.check_for_exception(self, ufw.common.UFWError, \
                                                   self.parser.parse_command,
                                                   c)

    def test_bad_simple_action(self):
        '''Test bad simple action'''
        for action in ['allw', 'eny', 'nonexistent']:
            c = ['rule', action, '22']
            tests.unit.support.check_for_exception(self, ValueError, \
                                                   self.parser.parse_command,
                                                   c)

    def test_delete_bad_simple_action(self):
        '''Test delete bad simple action'''
        for action in ['allw', 'eny', 'nonexistent']:
            c = ['rule', 'delete', action, '22']
            tests.unit.support.check_for_exception(self, ValueError, \
                                                   self.parser.parse_command,
                                                   c)

    def test_bad_simple_action_with_direction(self):
        '''Test bad simple action with direction'''
        for dir in ['ina', 'ou']:
            c = ['rule', 'allow', dir, '22']
            #self.parser.parse_command(c)
            tests.unit.support.check_for_exception(self, ufw.common.UFWError, \
                                                   self.parser.parse_command,
                                                   c)

        c = ['rule', 'allow', 5, '22']
        tests.unit.support.check_for_exception(self, AttributeError, \
                                               self.parser.parse_command,
                                               c)

    def test_route_delete_num(self):
        '''Test route delete NUM'''
        c = ['route', 'delete', '1']
        tests.unit.support.check_for_exception(self, ufw.common.UFWError, \
                                               self.parser.parse_command,
                                               c)

    def test_app_parse(self):
        '''Test UFWCommandApp.parse()'''
        cmds = [
                (['app', 'list'], None),
                (['app', 'info', 'WWW'], None),
                (['app', 'info', 'WWW Full'], None),
                (['app', 'info', 'Samba'], None),
                (['app', 'info', 'DNS'], None),
                (['app', 'update', 'WWW'], None),
                (['app', 'update', '--add-new', 'WWW'], None),
                (['app', 'default', 'allow'], None),
                (['app', 'default', 'deny'], None),
                (['app', 'default', 'reject'], None),
                (['app', 'default', 'skip'], None),
                (['notapp'], ValueError),
                (['app', 'default'], ValueError),
                (['app', 'list', 'extra args'], ValueError),
                (['app', 'info'], ValueError),
                (['app', 'default'], ValueError),
                (['app', 'default', 'nonexistent'], ValueError),
               ]
        for cmd, exception in cmds:
            #print(" ".join(cmd))

            if exception is not None:
                c = ufw.parser.UFWCommandApp(" ".join(cmd))
                tests.unit.support.check_for_exception(self, exception,
                                                       c.parse,
                                                       cmd + [])
            else:
                # Note, parser.parse_command() modifies its arg, so pass a copy
                # of the cmd, not a reference
                pr = self.parser.parse_command(cmd + [])

                # TODO: more tests here by sending the cmd and the pr to a
                # helper
                action = cmd[1]
                if action == 'update' and cmd[2] == '--add-new':
                    action = 'update-with-new'
                elif action == 'default':
                    action = "default-%s" % cmd[2]
                self.assertEquals(action, pr.action, "%s != %s" % (action, \
                                                                   pr.action))

    def test_default_parse(self):
        '''Test UFWCommandDefault.parse()'''
        cmds = [
                (['default', 'reject'], None),
                (['default', 'deny', 'incoming'], None),
                (['default', 'allow', 'outgoing'], None),
                (['default', 'deny', 'routed'], None),
                (['default'], ValueError),
                (['default', 'nonexistent'], ValueError),
                (['default', 'nonexistent', 'allow'], ValueError),
                (['default', 'incoming', 'allow'], ValueError),
                (['default', 'routed', 'deny'], ValueError),
               ]
        for cmd, exception in cmds:
            #print(" ".join(cmd))

            if exception is not None:
                c = ufw.parser.UFWCommandDefault(" ".join(cmd))
                tests.unit.support.check_for_exception(self, exception,
                                                       c.parse,
                                                       cmd + [])
            else:
                # Note, parser.parse_command() modifies its arg, so pass a copy
                # of the cmd, not a reference
                pr = self.parser.parse_command(cmd + [])

                # TODO: more tests here by sending the cmd and the pr to a
                # helper
                action = cmd[1]
                pol = "incoming"
                if len(cmd) >= 3:
                    pol = cmd[2]
                action = "default-%s-%s" % (cmd[1], pol)
                self.assertEquals(action, pr.action, "%s != %s" % (action, \
                                                                   pr.action))

    def test_logging_parse(self):
        '''Test UFWCommandLogging.parse()'''
        cmds = [
                (['logging', 'on'], None),
                (['logging', 'off'], None),
                (['logging', 'low'], None),
                (['logging', 'medium'], None),
                (['logging', 'high'], None),
                (['logging', 'full'], None),
                (['logging'], ValueError),
                (['logging', 'nonexistent'], ValueError),
               ]
        for cmd, exception in cmds:
            #print(" ".join(cmd))

            if exception is not None:
                c = ufw.parser.UFWCommandLogging(" ".join(cmd))
                tests.unit.support.check_for_exception(self, exception,
                                                       c.parse,
                                                       cmd + [])
            else:
                # Note, parser.parse_command() modifies its arg, so pass a copy
                # of the cmd, not a reference
                pr = self.parser.parse_command(cmd + [])

                # TODO: more tests here by sending the cmd and the pr to a
                # helper
                action = "logging-%s" % (cmd[1])
                if cmd[1] != "on" and cmd[1] != "off":
                    action = "logging-on_%s" % (cmd[1])
                self.assertEquals(action, pr.action, "%s != %s" % (action, \
                                                                   pr.action))

    def test_status_parse(self):
        '''Test UFWCommandStatus.parse()'''
        cmds = [
                (['status'], None),
                (['status', 'verbose'], None),
                (['status', 'numbered'], None),
                (['status', 'bad'], ValueError),
               ]
        for cmd, exception in cmds:
            #print(" ".join(cmd))

            if exception is not None:
                c = ufw.parser.UFWCommandStatus(" ".join(cmd))
                tests.unit.support.check_for_exception(self, exception,
                                                       c.parse,
                                                       cmd + [])
            else:
                # Note, parser.parse_command() modifies its arg, so pass a copy
                # of the cmd, not a reference
                pr = self.parser.parse_command(cmd + [])

                # TODO: more tests here by sending the cmd and the pr to a
                # helper
                action = cmd[0]
                if len(cmd) > 1:
                    action = "%s-%s" % (cmd[0], cmd[1])
                self.assertEquals(action, pr.action, "%s != %s" % (action, \
                                                                   pr.action))

    def test_show_parse(self):
        '''Test UFWCommandShow.parse()'''
        cmds = [
                (['show', 'raw'], None, "show-raw"),
                (['show', 'before-rules'], None, "show-before"),
                (['show', 'after-rules'], None, "show-after"),
                (['show', 'user-rules'], None, "show-user"),
                (['show', 'logging-rules'], None, "show-logging"),
                (['show', 'builtins'], None, "show-builtins"),
                (['show', 'listening'], None, "show-listening"),
                (['show', 'added'], None, "show-added"),
                (['show'], ValueError, None),
                (['show', 'bad'], ValueError, None),
               ]
        for cmd, exception, action in cmds:
            #print(" ".join(cmd))

            if exception is not None:
                c = ufw.parser.UFWCommandShow(" ".join(cmd))
                tests.unit.support.check_for_exception(self, exception,
                                                       c.parse,
                                                       cmd + [])
            else:
                # Note, parser.parse_command() modifies its arg, so pass a copy
                # of the cmd, not a reference
                pr = self.parser.parse_command(cmd + [])

                # TODO: more tests here by sending the cmd and the pr to a
                # helper
                self.assertEquals(action, pr.action, "%s != %s" % (action, \
                                                                   pr.action))

def test_main(): # used by runner.py
    tests.unit.support.run_unittest(
            ParserTestCase
    )

if __name__ == "__main__": # used when standalone
    unittest.main()



#!/usr/bin/python
#
# Copyright 2012 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# find_tests(), runtest() and main() inspired by regrtest.py from Python 3.1
# Copyright (c) 2001-2010 Python Software Foundation; All Rights Reserved
#


from __future__ import print_function
import os
import sys

def find_tests(testdir=None, testscripts=[]):
    '''Find tests'''
    if not testdir:
        if __name__ == '__main__':
            fn = sys.argv[0]
        else:
            print("TODO: find_tests() when imported")
            sys.exit(1)

        testdir = os.path.dirname(fn)

    if len(testscripts) > 1:
        names = testscripts[1:]
    else:
        names = os.listdir(testdir)
    tests = []
    for name in names:
        if name[:5] == "test_" and name[-3:] == ".py":
            tests.append(name[:-3])
    tests.sort()
    return tests

def runtest(test):
    '''Run test'''
    pkg =  __import__("tests.unit." + test, globals(), locals(), [])
    unit_pkg = getattr(pkg, "unit")
    mod = getattr(unit_pkg, test)
    print(test)
    mod.test_main()


if __name__ == '__main__':
    # Create the unittest symlink so imports work
    if not os.path.islink("./ufw"):
        os.symlink("./src", "./ufw")

    # Replace runner.py's directory from the search path, and add our own
    # so we can properly namespace our modules
    d = os.path.abspath(os.path.normpath(os.path.dirname(sys.argv[0])))
    testdir = os.path.dirname(d)
    testdir = os.path.dirname(os.path.dirname(d))
    i = len(sys.path)
    while i >= 0:
        i -= 1
        if os.path.abspath(os.path.normpath(sys.path[i])) == d:
            sys.path[i] = testdir

    print("DEBUG: sys.path=%s" % sys.path)
    tests = find_tests(testscripts=sys.argv)
    print("DEBUG: test=%s" % str(tests))

    # Import this here, so we are guaranteed to get ours from topdir
    from tests.unit.support import TestFailed

    passed = []
    failed = []
    skipped = []
    for test in tests:
        try:
            runtest(test)
            passed.append(test)
        except KeyboardInterrupt: # kill this test, but still do others
            print("")
            break
        except TestFailed as e:
            failed.append(test)
        except:
            raise

        # cleanup
        for m in list(sys.modules.keys()):
            if m.startswith("tests.unit.") and m != "tests.unit.support":
                try:
                    del sys.modules[m]
                except KeyError:
                    pass

    # Cleanup our symlink
    if os.path.islink("./ufw"):
        os.unlink("./ufw")

    print("")
    print("------------------")
    print("Unit tests summary")
    print("------------------")
    print("Total=%d (Passed=%d, Failed=%d)" % (len(passed) + len(failed),
                                               len(passed),
                                               len(failed)))
    if len(failed) > 0:
        sys.exit(1)



#
# Copyright 2012-2013 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# NOTE: most of this is tested via test_backend_iptables.py since we have to
#       have a backend defined to test backend.py

import unittest
import tests.unit.support
import ufw.backend
import ufw.backend_iptables
import ufw.common

class BackendTestCase(unittest.TestCase):
    def setUp(self):
        ufw.common.do_checks = False
        # ufw.backend is an interface, so to test it we need to instantiate
        # an object that implements this interface
        self.backend = ufw.backend_iptables.UFWBackendIptables(dryrun=True)

    def tearDown(self):
        pass

    def test_installation_defaults(self):
        '''Test installation defaults'''
        self.assertEquals(self.backend.defaults['default_input_policy'],
                          "drop")
        self.assertEquals(self.backend.defaults['default_forward_policy'],
                          "drop")
        self.assertEquals(self.backend.defaults['default_output_policy'],
                          "accept")
        self.assertEquals(self.backend.defaults['ipt_modules'],
                        'nf_conntrack_ftp nf_nat_ftp nf_conntrack_netbios_ns')
        self.assertEquals(self.backend.defaults['loglevel'], 'low')
        self.assertEquals(self.backend.defaults['manage_builtins'], 'no')
        self.assertEquals(self.backend.defaults['enabled'], 'no')
        self.assertEquals(self.backend.defaults['ipv6'], 'yes')
        self.assertEquals(self.backend.defaults['default_application_policy'],
                          'skip')


def test_main(): # used by runner.py
    tests.unit.support.run_unittest(
            BackendTestCase
    )

if __name__ == "__main__": # used when standalone
    unittest.main()



#
# Copyright 2012-2013 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import unittest
import tests.unit.support
import ufw.backend_iptables
import ufw.common
import ufw.frontend
import os
import re
import shutil
import time

try: # python 2
    from StringIO import StringIO
except ImportError:
    from io import StringIO

class BackendIptablesTestCase(unittest.TestCase):
    def setUp(self):
        ufw.common.do_checks = False

        for d in [ufw.common.state_dir, ufw.common.config_dir]:
            if not os.path.isdir(d + ".bak"):
                shutil.copytree(d, d + ".bak")

        # don't duplicate all the code for set_rule() from frontend.py so
        # the frontend's set_rule() to exercise our set_rule()
        self.ui = ufw.frontend.UFWFrontend(dryrun=True)

        # for convenience
        self.backend = self.ui.backend

        self.saved_msg_output = ufw.util.msg_output
        self.msg_output = None

        self.prevpath = os.environ['PATH']
        os.environ['PATH'] = "%s:%s" % (ufw.common.iptables_dir,
                                        os.environ['PATH'])

        # update ufw-init-functions to use our fake iptables* commands
        f = os.path.join(ufw.common.state_dir, "ufw-init-functions")
        contents = ""
        for line in open(f).readlines():
            if re.search("^PATH=", line):
                line = "#" + line
                line += 'PATH="%s:%s"\n' % (ufw.common.iptables_dir,
                                            line.split('"')[1])
            contents += line
        open(f + '.new', 'w').write(contents)
        os.rename(f + '.new', f)

    def tearDown(self):
        self.ui = None
        self.backend = None
        os.environ['PATH'] = self.prevpath

        for d in [ufw.common.state_dir, ufw.common.config_dir]:
            if os.path.isdir(d):
                tests.unit.support.recursive_rm(d)
                shutil.copytree(d + ".bak", d)

        if self.msg_output:
            ufw.util.msg_output = self.saved_msg_output
            self.msg_output.close()
            self.msg_output = None

        sysctl = os.path.join(ufw.common.iptables_dir, "sysctl")
        if os.path.exists(sysctl):
            os.unlink(sysctl)

    def _update_sysctl(self, forward=False):
        sysctl = os.path.join(ufw.common.iptables_dir, "sysctl")
        if forward:
            shutil.copy(os.path.join(ufw.common.iptables_dir,
                                     "sysctl-forward-yes"),
                        sysctl)
        else:
            shutil.copy(os.path.join(ufw.common.iptables_dir,
                                     "sysctl-forward-no"),
                        sysctl)

    def _test__do_checks(self):
        '''Test _do_checks()'''
        print ("  setting self.backend.do_checks to 'True'")
        self.backend.do_checks = True
        tests.unit.support.check_for_exception(self,
                              ufw.common.UFWError,
                              self.backend._do_checks)
        print ("  setting self.backend.do_checks to 'False'")
        self.backend.do_checks = False
        self.backend._do_checks()

    def test_get_default_application_policy(self):
        '''Test get_default_application_policy()'''
        s = self.backend.get_default_application_policy()
        self.assertTrue(s.endswith("skip"))

    def test_set_default_application_policy(self):
        '''Test set_default_application_policy()'''
        self.backend.dryrun = False
        for policy in ['allow', 'deny', 'reject', 'skip']:
            s = self.backend.set_default_application_policy(policy)
            self.assertTrue(policy in s, "Could not find '%s' in:\n%s" % \
                                         (policy, s))

    def test_get_app_rules_from_template(self):
        '''Test get_app_rules_from_template()'''
        pr = ufw.frontend.parse_command(['rule', 'allow', 'CIFS'])
        rules = self.backend.get_app_rules_from_template(pr.data['rule'])
        self.assertEquals(len(rules), 2)
        for r in rules:
            self.assertEquals(r.dapp, 'CIFS')

        pr = ufw.frontend.parse_command(['rule', 'deny',
                                         'from', 'any', 'app', 'CIFS'])
        rules = self.backend.get_app_rules_from_template(pr.data['rule'])
        self.assertEquals(len(rules), 2)
        for r in rules:
            self.assertEquals(r.sapp, 'CIFS')

        pr = ufw.frontend.parse_command(['rule', 'reject',
                                         'to', 'any', 'app', 'CIFS',
                                         'from', 'any', 'app', 'CIFS'])
        rules = self.backend.get_app_rules_from_template(pr.data['rule'])
        self.assertEquals(len(rules), 2)
        for r in rules:
            self.assertEquals(r.dapp, 'CIFS')
            self.assertEquals(r.sapp, 'CIFS')

        pr = ufw.frontend.parse_command(['rule', 'reject',
                                         'to', 'any', 'app', 'WWW',
                                         'from', 'any', 'app', 'WWW Secure'])
        rules = self.backend.get_app_rules_from_template(pr.data['rule'])
        self.assertEquals(len(rules), 1)
        for r in rules:
            self.assertEquals(r.dapp, 'WWW')
            self.assertEquals(r.sapp, 'WWW Secure')

        pr = ufw.frontend.parse_command(['rule', 'allow',
                                         'from', 'any', 'app', 'IPP',
                                         'to', 'any', 'app', 'WWW'])
        rules = self.backend.get_app_rules_from_template(pr.data['rule'])
        self.assertEquals(len(rules), 1)
        for r in rules:
            self.assertEquals(r.sapp, 'IPP')


        pr = ufw.frontend.parse_command(['rule', 'allow', '12345'])
        tests.unit.support.check_for_exception(self,
                              ufw.common.UFWError,
                              self.backend.get_app_rules_from_template,
                              pr.data['rule'])

    def test_update_app_rule(self):
        '''Test upate_app_rule()'''
        self.saved_msg_output = ufw.util.msg_output
        self.msg_output = StringIO()
        ufw.util.msg_output = self.msg_output

        (s, res) = self.backend.update_app_rule('WWW')
        self.assertFalse(res)
        self.assertEquals(s, "")

        pr = ufw.frontend.parse_command([] + ['rule', 'allow', 'CIFS'])
        self.backend.rules.append(pr.data['rule'])
        (s, res) = self.backend.update_app_rule('WWW')
        self.assertFalse(res)
        self.assertEquals(s, "")
        (s, res) = self.backend.update_app_rule('CIFS')
        self.assertTrue(res)
        self.assertTrue('CIFS' in s)

        pr = ufw.frontend.parse_command([] + ['rule', 'allow',
                                         'to', '5678:fff::/64',
                                         'app', 'WWW Secure'])
        pr.data['rule'].set_v6(True)
        self.backend.rules6.append(pr.data['rule'])
        (s, res) = self.backend.update_app_rule('WWW')
        self.assertFalse(res)
        self.assertEquals(s, "")
        (s, res) = self.backend.update_app_rule('WWW Secure')
        self.assertTrue(res)
        self.assertTrue('WWW Secure' in s)

        pr = ufw.frontend.parse_command([] + ['rule', 'allow',
                                         'from', '1234:fff::/64',
                                         'app', 'WWW Secure',
                                         'to', '2345:fff::/64',
                                         'app', 'WWW Full'])
        pr.data['rule'].set_v6(True)
        self.backend.rules6.append(pr.data['rule'])
        (s, res) = self.backend.update_app_rule('WWW')
        self.assertFalse(res)
        self.assertEquals(s, "")
        (s, res) = self.backend.update_app_rule('WWW Full')
        self.assertTrue(res)
        self.assertTrue('WWW Full' in s)

        pr = ufw.frontend.parse_command([] + ['rule', 'allow', 'NFS'])
        self.backend.rules.append(pr.data['rule'])
        pr.data['rule'].set_v6(True)
        self.backend.rules6.append(pr.data['rule'])
        (s, res) = self.backend.update_app_rule('WWW')
        self.assertFalse(res)
        self.assertEquals(s, "")
        (s, res) = self.backend.update_app_rule('NFS')
        self.assertTrue(res)
        self.assertTrue('NFS' in s)

    def test_find_application_name(self):
        '''Test find_application_name()'''
        res = self.backend.find_application_name('WWW')
        self.assertEquals(res, 'WWW')

        res = self.backend.find_application_name('WwW')
        self.assertEquals(res, 'WWW')

        f = os.path.join(self.backend.files['apps'], "testapp")
        contents = '''
[WWw]
title=Duplicate Web Server
description=Duplicate Web server
ports=80/tcp
'''
        fd = open(f, 'w')
        fd.write(contents)
        fd.close()
        self.backend.profiles = ufw.applications.get_profiles(
                                    self.backend.files['apps'])
        tests.unit.support.check_for_exception(self,
                              ufw.common.UFWError,
                              self.backend.find_application_name,
                              'wWw')

        tests.unit.support.check_for_exception(self,
                              ufw.common.UFWError,
                              self.backend.find_application_name,
                              'nonexistent')

    def test_find_other_position(self):
        '''Test find_other_position()'''
        pr = ufw.frontend.parse_command([] + ['rule', 'allow',
                                         'from', '1234:fff::/64',
                                         'app', 'WWW Secure',
                                         'to', '2345:fff::/64',
                                         'app', 'WWW Full'])
        pr.data['rule'].set_v6(True)
        self.backend.rules6.append(pr.data['rule'])

        pr = ufw.frontend.parse_command(['rule', 'allow', 'WWW'])
        self.backend.rules.append(pr.data['rule'])
        pr.data['rule'].set_v6(True)
        self.backend.rules6.append(pr.data['rule'])

        res = self.backend.find_other_position(2, v6=True)
        self.assertEquals(res, 0)


        res = self.backend.find_other_position(1, v6=False)
        self.assertEquals(res, 2)

        tests.unit.support.check_for_exception(self,
                              ValueError,
                              self.backend.find_other_position,
                              3,
                              True)

        tests.unit.support.check_for_exception(self,
                              ValueError,
                              self.backend.find_other_position,
                              3,
                              False)

        tests.unit.support.check_for_exception(self,
                              ValueError,
                              self.backend.find_other_position,
                              0,
                              False)

        pr = ufw.frontend.parse_command([] + ['rule', 'allow',
                                         'to', '2345:fff::/64',
                                         'app', 'CIFS'])
        pr.data['rule'].set_v6(True)
        self.backend.rules6.append(pr.data['rule'])

        pr = ufw.frontend.parse_command(['rule', 'allow', 'CIFS'])
        self.backend.rules.append(pr.data['rule'])
        pr.data['rule'].set_v6(True)
        self.backend.rules6.append(pr.data['rule'])

        res = self.backend.find_other_position(3, v6=True)
        self.assertEquals(res, 0)

    def test_get_loglevel(self):
        '''Test get_loglevel()'''
        for l in ['off', 'low', 'medium', 'high']:
            self.backend.set_loglevel(l)
            (level, s) = self.backend.get_loglevel()
            self.assertTrue(l in s, "Could not find '%s' in:\n%s" % (l, s))

        self.backend.defaults['loglevel'] = 'nonexistent'
        (level, s) = self.backend.get_loglevel()
        self.assertTrue('unknown' in s, "Could not find 'unknown' in:\n%s" % s)

    def test_set_loglevel(self):
        '''Test set_loglevel()'''
        for l in ['off', 'on', 'low', 'medium', 'high']:
            self.backend.set_loglevel(l)
            (level, s) = self.backend.get_loglevel()
            if l == 'on':
                l = 'low'
            self.assertTrue(l in s, "Could not find '%s' in:\n%s" % (l, s))

        tests.unit.support.check_for_exception(self,
                              ufw.common.UFWError,
                              self.backend.set_loglevel,
                              'nonexistent')

    def test_get_rules_count(self):
        '''Test get_rules_count()'''
        res = self.backend.get_rules_count(v6=False)
        self.assertEquals(res, 0)

        pr = ufw.frontend.parse_command([] + ['rule', 'allow',
                                         'from', '1234:fff::/64',
                                         'app', 'WWW Secure',
                                         'to', '2345:fff::/64',
                                         'app', 'WWW Full'])
        pr.data['rule'].set_v6(True)
        self.backend.rules6.append(pr.data['rule'])

        pr = ufw.frontend.parse_command(['rule', 'allow', 'WWW'])
        self.backend.rules.append(pr.data['rule'])
        pr.data['rule'].set_v6(True)
        self.backend.rules6.append(pr.data['rule'])

        res = self.backend.get_rules_count(v6=False)
        self.assertEquals(res, 1)

        res = self.backend.get_rules_count(v6=True)
        self.assertEquals(res, 2)

    def test_get_rule_by_number(self):
        '''Test get_rule_by_number()'''
        pr1 = ufw.frontend.parse_command(['rule', 'allow', 'WWW'])
        self.backend.rules.append(pr1.data['rule'])

        pr2 = ufw.frontend.parse_command(['rule', 'allow', 'WWW'])
        pr2.data['rule'].set_v6(True)
        self.backend.rules6.append(pr2.data['rule'])

        pr3 = ufw.frontend.parse_command([] + ['rule', 'allow',
                                         'from', '1234:fff::/64',
                                         'app', 'WWW Secure',
                                         'to', '2345:fff::/64',
                                         'app', 'WWW Full'])
        pr3.data['rule'].set_v6(True)
        self.backend.rules6.append(pr3.data['rule'])

        res = self.backend.get_rule_by_number(1)
        self.assertEquals(ufw.common.UFWRule.match(res, pr1.data['rule']), 0)
        self.assertEquals(ufw.common.UFWRule.match(res, pr2.data['rule']), 1)
        self.assertEquals(ufw.common.UFWRule.match(res, pr3.data['rule']), 1)

        res = self.backend.get_rule_by_number(2)
        self.assertEquals(ufw.common.UFWRule.match(res, pr2.data['rule']), 0)
        self.assertEquals(ufw.common.UFWRule.match(res, pr1.data['rule']), 1)
        self.assertEquals(ufw.common.UFWRule.match(res, pr3.data['rule']), 1)

        res = self.backend.get_rule_by_number(3)
        self.assertEquals(ufw.common.UFWRule.match(res, pr3.data['rule']), 0)
        self.assertEquals(ufw.common.UFWRule.match(res, pr1.data['rule']), 1)
        self.assertEquals(ufw.common.UFWRule.match(res, pr2.data['rule']), 1)

        res = self.backend.get_rule_by_number(4)
        self.assertEquals(res, None)

        pr4 = ufw.frontend.parse_command([] + ['rule', 'allow', 'CIFS'])
        self.backend.rules.append(pr4.data['rule'])
        pr4.data['rule'].set_v6(True)
        self.backend.rules6.append(pr4.data['rule'])
        res = self.backend.get_rule_by_number(6)
        self.assertEquals(res, None)
        res = self.backend.get_rule_by_number(4)
        self.assertEquals(ufw.common.UFWRule.match(res, pr4.data['rule']), 1)

    def test_get_matching(self):
        '''Test get_matching()'''
        pr1 = ufw.frontend.parse_command(['rule', 'allow', 'WWW'])
        self.backend.rules.append(pr1.data['rule'])

        pr2 = ufw.frontend.parse_command(['rule', 'deny', 'WWW'])
        self.backend.rules.append(pr2.data['rule'])

        test_rule = pr1.data['rule'].dup_rule()
        res = self.backend.get_matching(test_rule)
        self.assertEquals(len(res), 2)

    def test_set_bad_default_application_policy(self):
        '''Test bad set_default_application_policy()'''
        self.backend.dryrun = False
        for policy in ['alow', 'deny 78&']:
            tests.unit.support.check_for_exception(self,
                                  ufw.common.UFWError,
                                  self.backend.set_default_application_policy,
                                  policy)

    def test_set_default_policy(self):
        '''Test set_default_policy()'''
        # dryrun
        for direction in ['incoming', 'outgoing', 'routed']:
            for policy in ['allow', 'deny', 'reject']:
                res = self.backend.set_default_policy(policy, direction)
                self.assertTrue(policy in res,
                                "Could not find '%s' in:\n%s" % (policy, res))
                self.assertTrue(direction in res,
                                "Could not find '%s' in:\n%s" % (direction,
                                                                 res))

        # no dryrun
        self.backend.dryrun = False
        for direction in ['incoming', 'outgoing']:
            for policy in ['allow', 'deny', 'reject']:
                res = self.backend.set_default_policy(policy, direction)
                self.assertTrue(policy in res,
                                "Could not find '%s' in:\n%s" % (policy, res))
                self.assertTrue(direction in res,
                                "Could not find '%s' in:\n%s" % (direction,
                                                                 res))
                if direction == 'incoming':
                    res = self.backend._get_default_policy("input")
                else:
                    res = self.backend._get_default_policy("output")
                self.assertEquals(res, policy)

        #  no dryrun for routed
        self.backend.dryrun = False
        for forward_enabled in [ False, True ]:
            self._update_sysctl(forward_enabled)
            direction = "routed"
            for policy in ['allow', 'deny', 'reject']:
                res = self.backend.set_default_policy(policy, direction)
                self.assertTrue(policy in res,
                                "Could not find '%s' in:\n%s" % (policy, res))
                self.assertTrue(direction in res,
                                "Could not find '%s' in:\n%s" % (direction,
                                                                 res))
                res = self.backend._get_default_policy("forward",
                                                       check_forward=True)
                if not forward_enabled:
                    policy = "disabled"
                self.assertEquals(res, policy)

    def test_set_default(self):
        '''Test set_default()'''
        self.backend.set_default(self.backend.files['defaults'],
                                 'NEW_INPUT_POLICY',
                                 'accept')
        self.assertEquals(self.backend.defaults['new_input_policy'], 'accept')

    def test_set_bad_default(self):
        '''Test bad set_default_policy()'''
        tests.unit.support.check_for_exception(self,
                                       ufw.common.UFWError,
                                       self.backend.set_default,
                                       self.backend.files['defaults'],
                                       'DEFAULT INPUT_POLICY',
                                       "accept")

        tests.unit.support.check_for_exception(self,
                                       ufw.common.UFWError,
                                       self.backend.set_default,
                                       self.backend.files['defaults'] + \
                                               ".nonexistent",
                                       'DEFAULT_INPUT_POLICY',
                                       "accept")

    def test_get_running_raw(self):
        '''Test get_running_raw()'''
        # dryrun
        for t in ['raw', 'builtins', 'before', 'user', 'after', 'logging']:
            res = self.backend.get_running_raw(t)
            for s in ['iptables', 'ip6tables']:
                self.assertTrue("Checking raw %s" % s in res,
                                "Could not find '%s' in:\n%s" % (s, res))

        # no dryrun
        self.backend.dryrun = False
        for t in ['raw', 'builtins', 'before', 'user', 'after', 'logging']:
            res = self.backend.get_running_raw(t)
            self.assertTrue(t in res, "Could not find '%s' in:\n%s" % \
                            (t, res))

    def test_get_status(self):
        '''Test get_status()'''
        # build up some rules
        cmds_sim = tests.unit.support.get_sample_rule_commands_simple()
        cmds_ext = tests.unit.support.get_sample_rule_commands_extended()

        for cmds in [cmds_sim, cmds_ext]:
            self.backend.rules = []
            self.backend.rules6 = []
            for cmd in cmds:
                pr = ufw.frontend.parse_command(cmd + [])
                action = cmd[1]
                self.assertEquals(action, pr.action, "%s != %s" % (action, \
                                                               pr.action))
                if 'rule' in pr.data:
                    if pr.data['rule'].v6:
                        self.backend.rules6.append(pr.data['rule'])
                    else:
                        self.backend.rules.append(pr.data['rule'])

            # dryrun
            self.backend.dryrun = True
            for v in [False, True]:
                for c in [False, True]:
                    res = self.backend.get_status(verbose=v, show_count=c)
                    for s in ['iptables', 'ip6tables']:
                        self.assertTrue("Checking %s" % s in res,
                                        "Could not find '%s' in:\n%s" % (s, res))

            # no dryrun
            self.backend.dryrun = False
            for v in [False, True]:
                for c in [False, True]:
                    res = self.backend.get_status(verbose=v, show_count=c)
                    terms = ['Status: active', 'To']
                    if v:
                        terms += ['Logging: on', 'Default: deny',
                                  'New profiles: skip']
                    if c:
                        terms += '[ 1] '
                    for search in terms:
                        self.assertTrue(search in res,
                                        "Could not find '%s' in:\n%s" % (search,
                                                                     res))


    def test_stop_firewall(self):
        '''Test stop_firewall()'''
        self.backend.stop_firewall()
        self.backend.dryrun = False
        self.backend.stop_firewall()
        # TODO: verify output

    def test_start_firewall(self):
        '''Test start_firewall()'''
        self.backend.start_firewall()
        self.backend.dryrun = False
        self.backend.start_firewall()
        # TODO: verify output

    def test__need_reload(self):
        '''Test _need_reload()'''
        for v6 in [False, True]:
            res = self.backend._need_reload(v6)
            self.backend.dryrun = False
            res = self.backend._need_reload(v6)
            self.assertFalse(res)
            # TODO: verify output

    def test__reload_user_rules(self):
        '''Test _reload_user_rules()'''
        self.backend.defaults['enabled'] = "no"
        self.backend._reload_user_rules()
        self.backend.dryrun = False
        self.backend.defaults['enabled'] = "yes"
        self.backend._reload_user_rules()
        # TODO: verify output

    def test_use_ipv6(self):
        '''Test use_ipv6()'''
        self.backend.defaults['ipv6'] = "yes"
        self.assertTrue(self.backend.use_ipv6())
        self.backend.defaults['ipv6'] = "no"
        self.assertFalse(self.backend.use_ipv6())

    def test__get_defaults(self):
        '''Test _get_defaults()'''
        self.backend._get_defaults()
        for k in ['ipt_modules',
                  'default_output_policy',
                  'default_input_policy',
                  'default_forward_policy',
                  'loglevel',
                  'manage_builtins',
                  'enabled',
                  'ipv6',
                  'default_application_policy']:
            self.assertTrue(k in self.backend.defaults, "Could not find '%s'" \
                                                        % k)

        # Installation defaults are tested elsewhere

        f = self.backend.files['defaults']
        contents = ""
        for line in open(f).readlines():
            if re.search("^DEFAULT_INPUT_POLICY=", line):
                line = "#" + line
            contents += line
        fd = open(f + '.new', 'w')
        fd.write(contents)
        fd.close()
        os.rename(f + '.new', f)

        tests.unit.support.check_for_exception(self,
                              ufw.common.UFWError,
                              self.backend._get_defaults)

        f = self.backend.files['defaults']
        contents = ""
        for line in open(f).readlines():
            if re.search("^#DEFAULT_INPUT_POLICY=", line):
                line = "DEFAULT_INPUT_POLICY=bad" + line
            contents += line
        fd = open(f + '.new', 'w')
        fd.write(contents)
        fd.close()
        os.rename(f + '.new', f)

        tests.unit.support.check_for_exception(self,
                              ufw.common.UFWError,
                              self.backend._get_defaults)

    def test_set_rule(self):
        '''Test set_rule()'''
        self.ui.backend.dryrun = False # keeps the verbosity down
        # TODO: optimize this. We don't need to hit the disk for all of these.
        #       maybe set enabled to 'yes' once for each branch
        self.ui.backend.defaults['enabled'] = "yes"
        cmds_sim = tests.unit.support.get_sample_rule_commands_simple()
        for cmd in cmds_sim:
            pr = ufw.frontend.parse_command(cmd + [])
            action = cmd[1]
            self.assertEquals(action, pr.action, "%s != %s" % (action, \
                                                               pr.action))
            if 'rule' in pr.data:
                self.ui.do_action(pr.action, pr.data['rule'], \
                                  pr.data['iptype'], True)
            # TODO: verify output

    def test_update_logging(self):
        '''Test update_logging()'''
        self.backend.defaults['enabled'] = "no"
        self.backend.dryrun = False
        for level in ['off', 'low', 'medium', 'high', 'full']:
            self.backend.defaults['enabled'] = "no"
            self.backend.update_logging(level)
            self.backend.defaults['enabled'] = "yes"
            self.backend.update_logging(level)
            # TODO: verify output

    def test_reset(self):
        '''Test reset()'''
        res = self.backend.reset()
        print (res)

        # we only have 1 second resolution on the backup, so sleep is needed
        time.sleep(1)

        self.backend.dryrun = False
        res = self.backend.reset()
        print (res)
        # TODO: verify output


def test_main(): # used by runner.py
    tests.unit.support.run_unittest(
            BackendIptablesTestCase
    )

if __name__ == "__main__": # used when standalone
    unittest.main()



'''backend.py: interface for ufw backends'''
#
# Copyright 2008-2013 Canonical Ltd.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License version 3,
#    as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import errno
import os
import pwd
import re
import stat
import sys
import ufw.util
from ufw.util import error, warn, debug, _findpath
from ufw.common import UFWError, UFWRule
import ufw.applications

class UFWBackend:
    '''Interface for backends'''
    def __init__(self, name, dryrun, extra_files=None,
                 rootdir=None, datadir=None):
        self.defaults = None
        self.name = name
        self.dryrun = dryrun
        self.rules = []
        self.rules6 = []

        p = _findpath(ufw.common.config_dir, datadir)
        self.files = {'defaults': os.path.join(p, 'default/ufw'),
                      'conf': os.path.join(p, 'ufw/ufw.conf'),
                      'apps': os.path.join(p, 'ufw/applications.d') }
        if extra_files != None:
            self.files.update(extra_files)

        self.loglevels = {'off':       0,
                          'low':     100,
                          'medium':  200,
                          'high':    300,
                          'full':    400 }

        self.do_checks = ufw.common.do_checks
        self._do_checks()
        self._get_defaults()
        self._read_rules()

        self.profiles = ufw.applications.get_profiles(self.files['apps'])

        self.iptables = os.path.join(ufw.common.iptables_dir, "iptables")
        self.iptables_restore = os.path.join(ufw.common.iptables_dir, \
                                             "iptables-restore")
        self.ip6tables = os.path.join(ufw.common.iptables_dir, "ip6tables")
        self.ip6tables_restore = os.path.join(ufw.common.iptables_dir, \
                                              "ip6tables-restore")

        try:
            self.iptables_version = ufw.util.get_iptables_version(self.iptables)
        except OSError: # pragma: no coverage
            err_msg = _("Couldn't determine iptables version")
            raise UFWError(err_msg)

        # Initialize via initcaps only when we need it (LP: #1044361)
        self.caps = None

    def initcaps(self):
        '''Initialize the capabilities database. This needs to be called
           before accessing the database.'''

        # Only initialize if not initialized already
        if self.caps != None:
            return

        self.caps = {}
        self.caps['limit'] = {}

        # Set defaults for dryrun, non-root, etc
        self.caps['limit']['4'] = True
        self.caps['limit']['6'] = False # historical default for the testsuite

        # Try to get capabilities from the running system if root
        if self.do_checks and os.getuid() == 0 and not self.dryrun: # pragma: no coverage
            # v4
            try:
                nf_caps = ufw.util.get_netfilter_capabilities(self.iptables)
            except OSError as e:
                error("initcaps\n%s" % e)
            if 'recent-set' in nf_caps and 'recent-update' in nf_caps:
                self.caps['limit']['4'] = True
            else:
                self.caps['limit']['4'] = False

            # v6 (skip capabilities check for ipv6 if ipv6 is disabled in ufw
            # because the system may not have ipv6 support (LP: #1039729)
            if self.use_ipv6():
                try:
                    nf_caps = ufw.util.get_netfilter_capabilities(self.ip6tables)
                except OSError as e:
                    error("initcaps\n%s" % e)
                if 'recent-set' in nf_caps and 'recent-update' in nf_caps:
                    self.caps['limit']['6'] = True
                else:
                    self.caps['limit']['6'] = False

    def is_enabled(self):
        '''Is firewall configured as enabled'''
        if 'enabled' in self.defaults and \
           self.defaults['enabled'] == 'yes':
            return True
        return False

    def use_ipv6(self):
        '''Is firewall configured to use IPv6'''
        if 'ipv6' in self.defaults and \
           self.defaults['ipv6'] == 'yes' and \
           os.path.exists("/proc/sys/net/ipv6"):
            return True
        return False

    def _get_default_policy(self, primary="input", check_forward=False):
        '''Get default policy for specified primary chain'''
        policy = "default_" + primary + "_policy"

        rstr = ""
        if self.defaults[policy] == "accept":
            rstr = "allow"
        elif self.defaults[policy] == "reject":
            rstr = "reject"
        else:
            rstr = "deny"

        if check_forward and primary == "forward":
            enabled = False
            err_msg = _("problem running sysctl")

            (rc, out) = ufw.util.cmd(['sysctl', 'net.ipv4.ip_forward'])
            if rc != 0: # pragma: no cover
                raise UFWError(err_msg)
            if '1' in out:
                enabled = True

            # IPv6 may be disabled, so ignore sysctl output
            if self.use_ipv6():
                (rc, out) = ufw.util.cmd(['sysctl', 'net.ipv6.conf.default.forwarding'])
                if rc == 0 and '1' in out:
                    enabled = True

                (rc, out) = ufw.util.cmd(['sysctl', 'net.ipv6.conf.all.forwarding'])
                if rc == 0 and '1' in out:
                    enabled = True

            if not enabled:
                rstr = "disabled"

        return rstr

    # Don't do coverage on this cause we don't run the unit tests as root
    def _do_checks(self): # pragma: no coverage
        '''Perform basic security checks:
        is setuid or setgid (for non-Linux systems)
        checks that script is owned by root
        checks that every component in absolute path are owned by root
        warn if script is group writable
        warn if part of script path is group writable

        Doing this at the beginning causes a race condition with later
        operations that don't do these checks.  However, if the user running
        this script is root, then need to be root to exploit the race
        condition (and you are hosed anyway...)
        '''

        if not self.do_checks:
            err_msg = _("Checks disabled")
            warn(err_msg)
            return True

        # Not needed on Linux, but who knows the places we will go...
        if os.getuid() != os.geteuid():
            err_msg = _("ERROR: this script should not be SUID")
            raise UFWError(err_msg)
        if os.getgid() != os.getegid():
            err_msg = _("ERROR: this script should not be SGID")
            raise UFWError(err_msg)

        uid = os.getuid()
        if uid != 0:
            err_msg = _("You need to be root to run this script")
            raise UFWError(err_msg)

        # Use these so we only warn once
        warned_world_write = {}
        warned_group_write = {}
        warned_owner = {}

        profiles = []
        if not os.path.isdir(self.files['apps']):
            warn_msg = _("'%s' does not exist") % (self.files['apps'])
            warn(warn_msg)
        else:
            pat = re.compile(r'^\.')
            for profile in os.listdir(self.files['apps']):
                if not pat.search(profile):
                    profiles.append(os.path.join(self.files['apps'], profile))

        for path in list(self.files.values()) + \
                [ os.path.abspath(sys.argv[0]) ] + \
                profiles:
            if not path.startswith('/'):
                path = "%s/%s" % (os.getcwd(), path)
            while True:
                debug("Checking " + path)
                if path == self.files['apps'] and \
                           not os.path.isdir(self.files['apps']):
                    break

                try:
                    statinfo = os.stat(path)
                    mode = statinfo[stat.ST_MODE]
                except OSError:
                    err_msg = _("Couldn't stat '%s'") % (path)
                    raise UFWError(err_msg)
                except Exception:
                    raise

                # snaps and clicks unpack to this, so handle it
                click_user = 'clickpkg'
                snap_user = 'snappypkg'
                is_unpack_user = False
                try:
                    if pwd.getpwuid(statinfo.st_uid)[0] == click_user or \
                            pwd.getpwuid(statinfo.st_uid)[0] == snap_user:
                        is_unpack_user = True
                except KeyError:
                    pass

                if statinfo.st_uid != 0 and not is_unpack_user and \
                        path not in warned_owner:
                    warn_msg = _("uid is %(uid)s but '%(path)s' is owned by " \
                                 "%(st_uid)s") % ({'uid': str(uid), \
                                               'path': path, \
                                               'st_uid': str(statinfo.st_uid)})
                    warn(warn_msg)
                    warned_owner[path] = True
                if mode & stat.S_IWOTH and path not in warned_world_write:
                    warn_msg = _("%s is world writable!") % (path)
                    warn(warn_msg)
                    warned_world_write[path] = True
                if mode & stat.S_IWGRP and path not in warned_group_write:
                    warn_msg = _("%s is group writable!") % (path)
                    warn(warn_msg)
                    warned_group_write[path] = True

                if path == "/":
                    break

                last_path = path
                path = os.path.dirname(path)
                if not path:
                    raise OSError(errno.ENOENT, \
                                  "Could not find parent for '%s'" % \
                                  (last_path))

        for f in self.files:
            if f != 'apps' and not os.path.isfile(self.files[f]):
                err_msg = _("'%(f)s' file '%(name)s' does not exist") % \
                           ({'f': f, 'name': self.files[f]})
                raise UFWError(err_msg)

    def _get_defaults(self):
        '''Get all settings from defaults file'''
        self.defaults = {}
        for f in [self.files['defaults'], self.files['conf']]:
            try:
                orig = ufw.util.open_file_read(f)
            except Exception: # pragma: no coverage
                err_msg = _("Couldn't open '%s' for reading") % (f)
                raise UFWError(err_msg)
            pat = re.compile(r'^\w+="?\w+"?')
            for line in orig:
                if pat.search(line):
                    tmp = re.split(r'=', line.strip())
                    self.defaults[tmp[0].lower()] = tmp[1].lower().strip('"\'')

            orig.close()

        # do some default policy sanity checking
        policies = ['accept', 'drop', 'reject']
        for c in [ 'input', 'output', 'forward' ]:
            if 'default_%s_policy' % (c) not in self.defaults:
                err_msg = _("Missing policy for '%s'" % (c))
                raise UFWError(err_msg)
            p = self.defaults['default_%s_policy' % (c)]
            if p not in policies:
                err_msg = _("Invalid policy '%(policy)s' for '%(chain)s'" % \
                            ({'policy': p, 'chain': c}))
                raise UFWError(err_msg)

    def set_default(self, fn, opt, value):
        '''Sets option in defaults file'''
        if not re.match(r'^[\w_]+$', opt):
            err_msg = _("Invalid option")
            raise UFWError(err_msg)

        # Perform this here so we can present a nice error to the user rather
        # than a traceback
        if not os.access(fn, os.W_OK):
            err_msg = _("'%s' is not writable" % (fn))
            raise UFWError(err_msg)

        fns = ufw.util.open_files(fn)
        fd = fns['tmp']

        found = False
        pat = re.compile(r'^' + opt + '=')
        for line in fns['orig']:
            if pat.search(line):
                ufw.util.write_to_file(fd, opt + "=" + value + "\n")
                found = True
            else:
                ufw.util.write_to_file(fd, line)

        # Add the entry if not found
        if not found:
            ufw.util.write_to_file(fd, opt + "=" + value + "\n")

        try:
            ufw.util.close_files(fns)
        except Exception: # pragma: no coverage
            raise

        # Now that the files are written out, update value in memory
        self.defaults[opt.lower()] = value.lower().strip('"\'')

    def set_default_application_policy(self, policy):
        '''Sets default application policy of firewall'''
        if not self.dryrun:
            if policy == "allow":
                self.set_default(self.files['defaults'], \
                                        "DEFAULT_APPLICATION_POLICY", \
                                        "\"ACCEPT\"")
            elif policy == "deny":
                self.set_default(self.files['defaults'], \
                                        "DEFAULT_APPLICATION_POLICY", \
                                        "\"DROP\"")
            elif policy == "reject":
                self.set_default(self.files['defaults'], \
                                        "DEFAULT_APPLICATION_POLICY", \
                                        "\"REJECT\"")
            elif policy == "skip":
                self.set_default(self.files['defaults'], \
                                        "DEFAULT_APPLICATION_POLICY", \
                                        "\"SKIP\"")
            else:
                err_msg = _("Unsupported policy '%s'") % (policy)
                raise UFWError(err_msg)

        rstr = _("Default application policy changed to '%s'") % (policy)

        return rstr

    def get_app_rules_from_template(self, template):
        '''Return a list of UFWRules based on the template rule'''
        rules = []
        profile_names = list(self.profiles.keys())

        if template.dport in profile_names and template.sport in profile_names:
            dports = ufw.applications.get_ports(self.profiles[template.dport])
            sports = ufw.applications.get_ports(self.profiles[template.sport])
            for i in dports:
                tmp = template.dup_rule()
                tmp.dapp = ""
                tmp.set_port("any", "src")
                (port, proto) = ufw.util.parse_port_proto(i)
                tmp.set_protocol(proto)
                tmp.set_port(port, "dst")

                tmp.dapp = template.dapp

                if template.dport == template.sport:
                    # Just use the same ports as dst for src when they are the
                    # same to avoid duplicate rules
                    tmp.sapp = ""
                    (port, proto) = ufw.util.parse_port_proto(i)
                    tmp.set_protocol(proto)
                    tmp.set_port(port, "src")

                    tmp.sapp = template.sapp
                    rules.append(tmp)
                else:
                    for j in sports:
                        rule = tmp.dup_rule()
                        rule.sapp = ""
                        (port, proto) = ufw.util.parse_port_proto(j)
                        rule.set_protocol(proto)
                        rule.set_port(port, "src")

                        if rule.protocol == "any":
                            rule.set_protocol(tmp.protocol)

                        rule.sapp = template.sapp
                        rules.append(rule)
        elif template.sport in profile_names:
            for p in ufw.applications.get_ports(self.profiles[template.sport]):
                rule = template.dup_rule()
                rule.sapp = ""
                (port, proto) = ufw.util.parse_port_proto(p)
                rule.set_protocol(proto)
                rule.set_port(port, "src")

                rule.sapp = template.sapp
                rules.append(rule)
        elif template.dport in profile_names:
            for p in ufw.applications.get_ports(self.profiles[template.dport]):
                rule = template.dup_rule()
                rule.dapp = ""
                (port, proto) = ufw.util.parse_port_proto(p)
                rule.set_protocol(proto)
                rule.set_port(port, "dst")

                rule.dapp = template.dapp
                rules.append(rule)

        if len(rules) < 1:
            err_msg = _("No rules found for application profile")
            raise UFWError(err_msg)

        return rules

    def update_app_rule(self, profile):
        '''Update rule for profile in place. Returns result string and bool
           on whether or not the profile is used in the current ruleset.
        '''
        updated_rules = []
        updated_rules6 = []
        last_tuple = ""
        rstr = ""
        updated_profile = False

        # Remember, self.rules is from user[6].rules, and not the running
        # firewall.
        for r in self.rules + self.rules6:
            if r.dapp == profile or r.sapp == profile:
                # We assume that the rules are in app rule order. Specifically,
                # if app rule has multiple rules, they are one after the other.
                # If the rule ordering changes, the below will have to change.
                tupl = r.get_app_tuple()
                if tupl == last_tuple:
                    # Skip the rule if seen this tuple already (ie, it is part
                    # of a known tuple).
                    continue
                else:
                    # Have a new tuple, so find and insert new app rules here
                    template = r.dup_rule()
                    template.set_protocol("any")
                    if template.dapp != "":
                        template.set_port(template.dapp, "dst")
                    if template.sapp != "":
                        template.set_port(template.sapp, "src")
                    new_app_rules = self.get_app_rules_from_template(\
                                          template)

                    for new_r in new_app_rules:
                        new_r.normalize()
                        if new_r.v6:
                            updated_rules6.append(new_r)
                        else:
                            updated_rules.append(new_r)

                    last_tuple = tupl
                    updated_profile = True
            else:
                if r.v6:
                    updated_rules6.append(r)
                else:
                    updated_rules.append(r)

        if updated_profile:
            self.rules = updated_rules
            self.rules6 = updated_rules6
            rstr += _("Rules updated for profile '%s'") % (profile)

            try:
                self._write_rules(False) # ipv4
                self._write_rules(True) # ipv6
            except Exception: # pragma: no coverage
                err_msg = _("Couldn't update application rules")
                raise UFWError(err_msg)

        return (rstr, updated_profile)

    def find_application_name(self, profile_name):
        '''Find the application profile name for profile_name'''
        if profile_name in self.profiles:
            return profile_name

        match = ""
        matches = 0
        for n in list(self.profiles.keys()):
            if n.lower() == profile_name.lower():
                match = n
                matches += 1

        debug_msg = "'%d' matches for '%s'" % (matches, profile_name)
        debug(debug_msg)
        if matches == 1:
            return match
        elif matches > 1:
            err_msg = _("Found multiple matches for '%s'. Please use exact profile name") % \
                        (profile_name)
        else:
            err_msg = _("Could not find a profile matching '%s'") % \
                        (profile_name)
        raise UFWError(err_msg)

    def find_other_position(self, position, v6):
        '''Return the absolute position in the other list of the rule with the
	   user position of the given list. For example, find_other_position(4,
	   True) will return the absolute position of the rule in the ipv4 list
           matching the user specified '4' rule in the ipv6 list.
        '''
        # Invalid search (v6 rule with too low position)
        if v6 and position > len(self.rules6):
            raise ValueError()

        # Invalid search (v4 rule with too high position)
        if not v6 and position > len(self.rules):
            raise ValueError()

        if position < 1:
            raise ValueError()

        rules = []
        if v6:
            rules = self.rules6
        else:
            rules = self.rules

        # self.rules[6] is a list of tuples. Some application rules have
        # multiple tuples but the user specifies by ufw rule, not application
        # tuple, so we need to find how many tuples there are leading up to
        # the specified position, which we can then use as an offset for
        # getting the proper match_rule.
        app_rules = {}
        tuple_offset = 0
        for i, r in enumerate(rules):
            if i >= position:
                break
            tupl = ""
            if r.dapp != "" or r.sapp != "":
                tupl = r.get_app_tuple()

                if tupl in app_rules:
                    tuple_offset += 1
                else:
                    app_rules[tupl] = True

        rules = []
        if v6:
            rules = self.rules
            match_rule = self.rules6[position - 1 + tuple_offset].dup_rule()
            match_rule.set_v6(False)
        else:
            rules = self.rules6
            match_rule = self.rules[position - 1 + tuple_offset].dup_rule()
            match_rule.set_v6(True)

        count = 1
        for r in rules:
            if UFWRule.match(r, match_rule) == 0:
                return count
            count += 1

        return 0

    def get_loglevel(self):
        '''Gets current log level of firewall'''
        level = 0
        rstr = _("Logging: ")
        if 'loglevel' not in self.defaults or \
           self.defaults['loglevel'] not in list(self.loglevels.keys()):
            level = -1
            rstr += _("unknown")
        else:
            level = self.loglevels[self.defaults['loglevel']]
            if level == 0:
                rstr += "off"
            else:
                rstr += "on (%s)" % (self.defaults['loglevel'])
        return (level, rstr)

    def set_loglevel(self, level):
        '''Sets log level of firewall'''
        if level not in list(self.loglevels.keys()) + ['on']:
            err_msg = _("Invalid log level '%s'") % (level)
            raise UFWError(err_msg)

        new_level = level
        if level == "on":
            if 'loglevel' not in self.defaults or \
               self.defaults['loglevel'] == "off":
                new_level = "low"
            else:
                new_level = self.defaults['loglevel']

        self.set_default(self.files['conf'], "LOGLEVEL", new_level)
        self.update_logging(new_level)

        if new_level == "off":
            return _("Logging disabled")
        else:
            return _("Logging enabled")

    def get_rules(self):
        '''Return list of all rules'''
        return self.rules + self.rules6

    def get_rules_count(self, v6):
        '''Return number of ufw rules (not iptables rules)'''
        rules = []
        if v6:
            rules = self.rules6
        else:
            rules = self.rules

        count = 0
        app_rules = {}
        for r in rules:
            tupl = ""
            if r.dapp != "" or r.sapp != "":
                tupl = r.get_app_tuple()

                if tupl in app_rules:
                    debug("Skipping found tuple '%s'" % (tupl))
                    continue
                else:
                    app_rules[tupl] = True
            count += 1

        return count

    def get_rule_by_number(self, num):
        '''Return rule specified by number seen via "status numbered"'''
        rules = self.get_rules()

        count = 1
        app_rules = {}
        for r in rules:
            tupl = ""
            if r.dapp != "" or r.sapp != "":
                tupl = r.get_app_tuple()

                if tupl in app_rules:
                    debug("Skipping found tuple '%s'" % (tupl))
                    continue
                else:
                    app_rules[tupl] = True
            if count == int(num):
                return r
            count += 1

        return None

    def get_matching(self, rule):
        '''See if there is a matching rule in the existing ruleset. Note this
           does not group rules by tuples.'''
        matched = []
        count = 0
        for r in self.get_rules():
            count += 1
            ret = rule.fuzzy_dst_match(r)
            if ret < 1:
                matched.append(count)

        return matched

    # API overrides
    def set_default_policy(self, policy, direction): # pragma: no coverage
        '''Set default policy for specified direction'''
        raise UFWError("UFWBackend.set_default_policy: need to override")

    def get_running_raw(self, rules_type): # pragma: no coverage
        '''Get status of running firewall'''
        raise UFWError("UFWBackend.get_running_raw: need to override")

    def get_status(self, verbose, show_count): # pragma: no coverage
        '''Get managed rules'''
        raise UFWError("UFWBackend.get_status: need to override")

    def set_rule(self, rule, allow_reload): # pragma: no coverage
        '''Update firewall with rule'''
        raise UFWError("UFWBackend.set_rule: need to override")

    def start_firewall(self): # pragma: no coverage
        '''Start the firewall'''
        raise UFWError("UFWBackend.start_firewall: need to override")

    def stop_firewall(self): # pragma: no coverage
        '''Stop the firewall'''
        raise UFWError("UFWBackend.stop_firewall: need to override")

    def get_app_rules_from_system(self, template, v6): # pragma: no coverage
        '''Get a list if rules based on template'''
        raise UFWError("UFWBackend.get_app_rules_from_system: need to " + \
                       "override")

    def update_logging(self, level): # pragma: no coverage
        '''Update loglevel of running firewall'''
        raise UFWError("UFWBackend.update_logging: need to override")

    def reset(self): # pragma: no coverage
        '''Reset the firewall'''
        raise UFWError("UFWBackend.reset: need to override")



#
# Copyright 2012-2016 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import unittest
import tests.unit.support
import ufw.common

class CommonTestCase(unittest.TestCase):
    def setUp(self):
        self.rules = {
                "any":  ufw.common.UFWRule("allow", "any"),
                "ipv6": ufw.common.UFWRule("deny", "ipv6"),
                "tcp":  ufw.common.UFWRule("limit", "tcp"),
                "udp":  ufw.common.UFWRule("allow", "udp"),
                "reject-tcp":  ufw.common.UFWRule("reject", "tcp"),
                "reject-udp":  ufw.common.UFWRule("reject", "udp"),
                "full-any":  ufw.common.UFWRule("allow", "any",
                    dport="123", dst="10.0.0.1", sport="124",
                    src="10.0.0.2", direction="in"),
                "full-ipv6": ufw.common.UFWRule("deny", "ipv6",
                    dport="123", dst="10.0.0.1", sport="124",
                    src="10.0.0.2", direction="in"),
                "full-tcp":  ufw.common.UFWRule("limit", "tcp",
                    dport="123", dst="10.0.0.1", sport="124",
                    src="10.0.0.2", direction="out"),
                "full-udp":  ufw.common.UFWRule("reject", "udp",
                    dport="123", dst="10.0.0.1", sport="124",
                    src="10.0.0.2", direction="out"),
                "dapp":  ufw.common.UFWRule("allow", "any"),
                "sapp":  ufw.common.UFWRule("deny", "any"),
                "app-both":  ufw.common.UFWRule("deny", "any"),
                "multi-dport": ufw.common.UFWRule("allow", "tcp",
                    dport="80,443,8080:8090"),
                "multi-sport": ufw.common.UFWRule("allow", "tcp",
                    sport="80,443,8080:8090"),
                "multi-both": ufw.common.UFWRule("allow", "tcp",
                    dport="80,443,8080:8090", sport="23"),
                "log":  ufw.common.UFWRule("allow", "tcp", dport="22"),
                "log-all":  ufw.common.UFWRule("allow", "tcp", dport="22"),
                }
        self.rules['dapp'].dapp = "Apache"
        self.rules['dapp'].dport = "80"
        self.rules['dapp'].proto = "tcp"

        self.rules['sapp'].sapp = "Apache"
        self.rules['sapp'].sport = "80"
        self.rules['sapp'].proto = "tcp"

        self.rules['app-both'].dapp = "Apache"
        self.rules['app-both'].dport = "80"
        self.rules['app-both'].proto = "tcp"
        self.rules['app-both'].sapp = "Apache"
        self.rules['app-both'].sport = "80"
        self.rules['app-both'].proto = "tcp"

        self.rules['log'].set_logtype("log")
        self.rules['log-all'].set_logtype("log-all")

    def tearDown(self):
        self.rules = None

    def test_ufwerror(self):
        '''Test UFWError'''
        try:
            raise ufw.common.UFWError("test")
        except ufw.common.UFWError as e:
            self.assertEquals(e.value, "test", "'%s' != 'test'" % e.value)
            return
        self.assertTrue(False, "Did not raise an error")

    def test_ufwerror_str(self):
        '''Test UFWError.str()'''
        e = ufw.common.UFWError("test")
        search = repr("test")
        self.assertEquals(str(e), search, "'%s' != 'test'" % search)

    def test__init_(self):
        '''Test UFWRule.__init__()'''
        r = ufw.common.UFWRule("allow", "tcp", "22")
        self.assertEquals(r.action, "allow")
        self.assertEquals(r.protocol, "tcp")
        self.assertEquals(r.dport, "22")

        tests.unit.support.check_for_exception(self, ufw.common.UFWError,
                                               ufw.common.UFWRule,
                                               "allow",
                                               "nonexistent",
                                               "22")

    def test__get_attrib(self):
        '''Test _get_attrib()'''
        res = self.rules["any"]._get_attrib()
        search = "'-p all -j ACCEPT', action=allow, comment=, dapp=, " + \
                 "direction=in, dport=any, dst=0.0.0.0/0, forward=False, " + \
                 "interface_in=, interface_out=, logtype=, multi=False, " + \
                 "position=0, protocol=any, remove=False, sapp=, " + \
                 "sport=any, src=0.0.0.0/0, updated=False, v6=False"
        self.assertEquals(res, search, "'%s' != '%s'" % (res, search))

    def test_dup_rule(self):
        '''Test dup_rule()'''
        r = self.rules["any"].dup_rule()
        self.assertEquals(ufw.common.UFWRule.match(r, self.rules["any"]), 0)

    def test_format_rule(self):
        '''Test format_rule()'''
        s = str(self.rules["any"])
        self.assertEquals(s, "-p all -j ACCEPT")

        s = str(self.rules["app-both"])
        self.assertEquals(s, "-p all --dport 80 --sport 80 -j DROP " + \
                             "-m comment --comment 'dapp_Apache,sapp_Apache'")

        s = str(self.rules["dapp"])
        self.assertEquals(s, "-p all --dport 80 -j ACCEPT " + \
                             "-m comment --comment 'dapp_Apache'")

        s = str(self.rules["full-any"])
        self.assertEquals(s, "-p all -d 10.0.0.1 --dport 123 " + \
                             "-s 10.0.0.2 --sport 124 -j ACCEPT")

        s = str(self.rules["full-ipv6"])
        self.assertEquals(s, "-p ipv6 -d 10.0.0.1 --dport 123 " + \
                             "-s 10.0.0.2 --sport 124 -j DROP")

        s = str(self.rules["full-tcp"])
        self.assertEquals(s, "-p tcp -d 10.0.0.1 --dport 123 " + \
                             "-s 10.0.0.2 --sport 124 -j LIMIT")

        s = str(self.rules["full-udp"])
        self.assertEquals(s, "-p udp -d 10.0.0.1 --dport 123 " + \
                             "-s 10.0.0.2 --sport 124 -j REJECT")

        s = str(self.rules["ipv6"])
        self.assertEquals(s, "-p ipv6 -j DROP")

        s = str(self.rules["log"])
        self.assertEquals(s, "-p tcp --dport 22 -j ACCEPT_log")

        s = str(self.rules["log-all"])
        self.assertEquals(s, "-p tcp --dport 22 -j ACCEPT_log-all")
        r = self.rules["log-all"].dup_rule()
        r.set_action("deny_log-all")
        s = str(r)
        self.assertEquals(s, "-p tcp --dport 22 -j DROP_log-all")

        s = str(self.rules["multi-both"])
        self.assertEquals(s, "-p tcp -m multiport " + \
                             "--dports 80,443,8080:8090 " + \
                             "-m multiport --sports 23 -j ACCEPT")

        s = str(self.rules["multi-dport"])
        self.assertEquals(s, "-p tcp -m multiport " + \
                             "--dports 80,443,8080:8090 -j ACCEPT")

        s = str(self.rules["multi-sport"])
        self.assertEquals(s, "-p tcp -m multiport " + \
                             "--sports 80,443,8080:8090 -j ACCEPT")

        s = str(self.rules["reject-tcp"])
        self.assertEquals(s, "-p tcp -j REJECT --reject-with tcp-reset")

        s = str(self.rules["reject-udp"])
        self.assertEquals(s, "-p udp -j REJECT")

        s = str(self.rules["sapp"])
        self.assertEquals(s, "-p all --sport 80 -j DROP " + \
                             "-m comment --comment 'sapp_Apache'")

        s = str(self.rules["tcp"])
        self.assertEquals(s, "-p tcp -j LIMIT")

        s = str(self.rules["udp"])
        self.assertEquals(s, "-p udp -j ACCEPT")

    def test_set_action(self):
        '''Test set_action()'''
        r = self.rules["any"]
        for action in ['allow', 'deny', 'reject', 'limit']:
            r.set_action(action)
            self.assertEquals(action, r.action, "%s != %s" %
                              (action, r.action))

    def test_set_port(self):
        '''Test set_port()'''
        rule = self.rules["any"]
        for loc in ['dst', 'src']:
            for port in ['any',
                         '1',
                         '22',
                         '1023',
                         '1024',
                         '1025',
                         '65535',
                         '1,2,3,4,5,6,7,8,9,10,11,12,13,14,15',
                         '80,443,8080:8090',
                         '22:25']:
                r = rule.dup_rule()
                r.set_port(port, loc)
                if loc == 'dst':
                    self.assertEquals(port, r.dport, "%s != %s" % (port,
                                                                   r.dport))
                else:
                    self.assertEquals(port, r.sport, "%s != %s" % (port,
                                                                   r.sport))

        r = self.rules["dapp"].dup_rule()
        r.dapp = "Apache"
        r.set_port("Apache", "dst")
        self.assertEquals(r.dapp, r.dport, "%s != %s" % (r.dapp, r.dport))

        r = self.rules["sapp"].dup_rule()
        r.sapp = "Apache"
        r.set_port("Apache", "src")
        self.assertEquals(r.sapp, r.sport, "%s != %s" % (r.sapp, r.sport))

    def test_set_port_bad(self):
        '''Test set_port() - bad'''
        rule = self.rules["any"]
        for loc in ['dst', 'src']:
            for port in ['an',
                         '0',
                         ',',
                         '',
                         ' ',
                         22,
                         '65536',
                         ',443,8080:8090',
                         '443:8080:8090',
                         '0:65536',
                         '2:1',
                         '80,',
                         '80,443,8080:',
                         ':8090',
                         '1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16',
                        ]:
                r = rule.dup_rule()
                e = ufw.common.UFWError
                if port == 22:
                    e = TypeError
                tests.unit.support.check_for_exception(self,
                                                       e,
                                                       r.set_port,
                                                       port,
                                                       loc)

    def test_set_protocol(self):
        '''Test set_protocol()'''
        r = self.rules["any"]
        for proto in ['any', 'tcp', 'udp', 'ipv6', 'esp', 'ah']:
            r.set_protocol(proto)
            self.assertEquals(proto, r.protocol, "%s != %s" %
                              (proto, r.protocol))

    def test_set_protocol_bad(self):
        '''Test set_protocol() - bad'''
        r = self.rules["any"]
        for proto in ['an', 'cp', 'up', 'nonexistent']:
            tests.unit.support.check_for_exception(self,
                                                   ufw.common.UFWError,
                                                   r.set_protocol,
                                                   proto)

    def test__fix_anywhere(self):
        '''Test _fix_anywhere()'''
        x = self.rules["any"].dup_rule()
        x.set_v6(False)
        x._fix_anywhere()
        search = "0.0.0.0/0"
        self.assertEquals(x.dst, search, "'%s' != '%s'" % (x.dst, search))

        y = x.dup_rule()
        y.set_v6(True)
        y._fix_anywhere()
        search = "::/0"
        self.assertEquals(y.dst, search, "'%s' != '%s'" % (y.dst, search))

    def test_set_v6(self):
        '''Test set_v6()'''
        r = self.rules["any"]
        for ipv6 in [True, False]:
            r.set_v6(ipv6)
            self.assertEquals(ipv6, r.v6, "%s != %s" %
                              (ipv6, r.v6))

    def test_set_src(self):
        '''Test set_src()'''
        r = self.rules["any"]
        for src in ["10.0.0.3"]:
            r.set_src(src)
            self.assertEquals(src, r.src, "%s != %s" %
                              (src, r.src))

    def test_set_src_bad(self):
        '''Test set_src() - bad'''
        r = self.rules["any"]
        for src in ["10.0.0.", "10..0.0.3"]:
            tests.unit.support.check_for_exception(self,
                                                   ufw.common.UFWError,
                                                   r.set_src,
                                                   src)

    def test_set_dst(self):
        '''Test set_dst()'''
        r = self.rules["any"]
        for dst in ["10.0.0.3"]:
            r.set_dst(dst)
            self.assertEquals(dst, r.dst, "%s != %s" %
                              (dst, r.dst))

    def test_set_dst_bad(self):
        '''Test set_dst() - bad'''
        r = self.rules["any"]
        for dst in ["10.0.0.", "10..0.0.3"]:
            tests.unit.support.check_for_exception(self,
                                                   ufw.common.UFWError,
                                                   r.set_dst,
                                                   dst)

    def test_set_interface(self):
        '''Test set_interface()'''
        r = self.rules["any"]
        for if_type in ["in", "out"]:
            for interface in ["eth0", "wlan1", "br_lan"]:
                r.set_interface(if_type, interface)
                if if_type == "in":
                    self.assertEquals(interface, r.interface_in, "%s != %s" %
                                      (interface, r.interface_in))
                else:
                    self.assertEquals(interface, r.interface_out, "%s != %s" %
                                      (interface, r.interface_out))

    def test_set_interface_bad(self):
        '''Test set_interface() - bad'''
        r = self.rules["any"]
        interface = "eth0"
        for if_type in ["ina", "ot"]:
            tests.unit.support.check_for_exception(self,
                                                   ufw.common.UFWError,
                                                   r.set_interface,
                                                   if_type,
                                                   interface)

        for if_type in ["in", "out"]:
            for interface in ["\tfoo", "<$%", "0eth", "eth0:0", "!eth0"]:
                tests.unit.support.check_for_exception(self,
                                                       ufw.common.UFWError,
                                                       r.set_interface,
                                                       if_type,
                                                       interface)

    def test_set_position(self):
        '''Test set_position()'''
        r = self.rules["any"]
        r.set_position(2)
        self.assertEquals(2, r.position)

    def test_set_position_bad(self):
        '''Test set_position() - bad'''
        r = self.rules["any"]
        tests.unit.support.check_for_exception(self,
                                               ufw.common.UFWError,
                                               r.set_position,
                                               'a')

    def test_set_logtype(self):
        '''Test set_logtype()'''
        r = self.rules["any"]
        for logtype in ["", "log", "log-all"]:
            r.set_logtype(logtype)
            self.assertEquals(logtype, r.logtype, "%s != %s" %
                              (logtype, r.logtype))

    def test_set_logtype_bad(self):
        '''Test set_logtype() - bad'''
        r = self.rules["any"]
        for logtype in ["a", "loga", "d"]:
            tests.unit.support.check_for_exception(self,
                                                   ufw.common.UFWError,
                                                   r.set_logtype,
                                                   logtype)

    def test_set_direction(self):
        '''Test set_direction()'''
        r = self.rules["any"]
        for direction in ["in", "out"]:
            r.set_direction(direction)
            self.assertEquals(direction, r.direction, "%s != %s" %
                              (direction, r.direction))

    def test_set_direction_bad(self):
        '''Test set_direction() - bad'''
        r = self.rules["any"]
        for direction in ["", "ina", "outta"]:
            tests.unit.support.check_for_exception(self,
                                                   ufw.common.UFWError,
                                                   r.set_direction,
                                                   direction)

    def test_normalize(self):
        '''Test normalize()'''
        # Test the pre-canned rules above-- none of them are normalized, so
        # UFWRule.match() should always be 0
        keys = list(self.rules.keys())
        keys.sort()
        for rule in keys:
            r = self.rules[rule].dup_rule()
            r.normalize()
            self.assertEquals(ufw.common.UFWRule.match(self.rules[rule], r), 0,
                            "'%s' != '%s'" % (self.rules[rule], r))

        # Bad rules
        bad = ufw.common.UFWRule("allow", "any")
        bad.src = "1000.0.0.1"
        tests.unit.support.check_for_exception(self,
                                               ufw.common.UFWError,
                                               bad.normalize)
        bad = None
        bad = ufw.common.UFWRule("allow", "any")
        bad.dst = "1000.0.0.1"
        tests.unit.support.check_for_exception(self,
                                               ufw.common.UFWError,
                                               bad.normalize)

        # Normalized rules
        data = [
                 (False, '192.168.0.1', '192.168.0.1'),
                 (False, '192.168.0.1/31', '192.168.0.0/31'),
                 (False, '192.168.0.1/255.255.255.0', '192.168.0.0/24'),
                 (True, '::1', '::1'),
                 (True, 'ff80:123:4567:89ab:cdef:123:4567:89ab/112',
                        'ff80:123:4567:89ab:cdef:123:4567:89ab/112')
                ]
        for (v6, addr, expected) in data:
            rule = ufw.common.UFWRule("allow", "any", dst=addr)
            rule.set_v6(v6)
            rule.normalize()
            self.assertEquals(expected, rule.dst,
                              "'%s' != '%s'" % (expected, rule.dst))
            self.assertEquals(addr != expected, rule.updated,
                              "'%s' not updated" % addr)

            rule = ufw.common.UFWRule("allow", "any", src=addr)
            rule.set_v6(v6)
            rule.normalize()
            self.assertEquals(expected, rule.src,
                              "'%s' != '%s'" % (expected, rule.src))
            self.assertEquals(addr != expected, rule.updated,
                              "'%s' not updated" % addr)

    def test_match(self):
        '''Test match()'''
        x = self.rules["full-any"].dup_rule()
        y = self.rules["full-any"].dup_rule()
        self.assertEquals(ufw.common.UFWRule.match(x, y), 0)

        for action in ['reject', 'deny', 'limit']:
            y = self.rules["full-any"].dup_rule()
            y.set_action(action)
            self.assertEquals(ufw.common.UFWRule.match(x, y), -1)

        for logtype in ['log', 'log-all']:
            y = self.rules["full-any"].dup_rule()
            y.set_logtype(logtype)
            self.assertEquals(ufw.common.UFWRule.match(x, y), -1)

        for comment in ['comment1', 'comment2']:
            y = self.rules["full-any"].dup_rule()
            y.set_comment(comment)
            self.assertEquals(ufw.common.UFWRule.match(x, y), -2)

        y = self.rules["full-any"].dup_rule()
        y.set_port("456", loc="dst")
        self.assertEquals(ufw.common.UFWRule.match(x, y), 1)

        y = self.rules["full-any"].dup_rule()
        y.set_port("456", loc="src")
        self.assertEquals(ufw.common.UFWRule.match(x, y), 1)

        y = self.rules["full-any"].dup_rule()
        y.set_protocol("tcp")
        self.assertEquals(ufw.common.UFWRule.match(x, y), 1)

        y = self.rules["full-any"].dup_rule()
        y.set_src("192.168.0.1")
        self.assertEquals(ufw.common.UFWRule.match(x, y), 1)

        y = self.rules["full-any"].dup_rule()
        y.set_dst("192.168.0.1")
        self.assertEquals(ufw.common.UFWRule.match(x, y), 1)

        y = self.rules["full-any"].dup_rule()
        y.set_dst("fe80::1")
        self.assertEquals(ufw.common.UFWRule.match(x, y), 1)

        y = ufw.common.UFWRule("allow", "tcp", dst="fe80::1")
        self.assertEquals(ufw.common.UFWRule.match(x, y), 1)

        y = self.rules["full-any"].dup_rule()
        y.sapp = "OpenSSH"
        self.assertEquals(ufw.common.UFWRule.match(x, y), 1)

        y = self.rules["full-any"].dup_rule()
        y.dapp = "OpenSSH"
        self.assertEquals(ufw.common.UFWRule.match(x, y), 1)

        y = self.rules["full-any"].dup_rule()
        y.set_interface("in", "eth0")
        self.assertEquals(ufw.common.UFWRule.match(x, y), 1)

        x = ufw.common.UFWRule("allow", "tcp", direction="out")
        y = x.dup_rule()
        y.set_interface("out", "eth0")
        self.assertEquals(ufw.common.UFWRule.match(x, y), 1)

        x = self.rules["any"].dup_rule()
        y = self.rules["any"].dup_rule()
        y.v6 = True
        self.assertEquals(ufw.common.UFWRule.match(x, y), 1)

        x = self.rules["any"].dup_rule()
        y = self.rules["any"].dup_rule()
        y.forward = True
        self.assertEquals(ufw.common.UFWRule.match(x, y), 1)

        x = self.rules["full-any"].dup_rule()
        y = self.rules["full-any"].dup_rule()
        y.forward = True
        self.assertEquals(ufw.common.UFWRule.match(x, y), 1)

        x = self.rules["multi-both"].dup_rule()
        y = self.rules["multi-both"].dup_rule()
        y.forward = True
        self.assertEquals(ufw.common.UFWRule.match(x, y), 1)

        x = ufw.common.UFWRule("allow", "tcp", direction="out")
        x.set_interface("out", "eth0")
        y = x.dup_rule()
        y.direction = "in"
        self.assertEquals(ufw.common.UFWRule.match(x, y), 1)

        tests.unit.support.check_for_exception(self, ValueError,
                                               x.match,
                                               None)

    def test_fuzzy_dst_match(self):
        '''Test fuzzy_dst_match()'''
        x = self.rules["full-any"].dup_rule()
        y = self.rules["full-any"].dup_rule()
        self.assertEquals(ufw.common.UFWRule.fuzzy_dst_match(x, y), 0)
        x.set_protocol("tcp")
        self.assertEquals(ufw.common.UFWRule.fuzzy_dst_match(x, y), -1)
        self.assertEquals(ufw.common.UFWRule.fuzzy_dst_match(y, x), 1)

        x = self.rules["multi-dport"].dup_rule()
        y = self.rules["multi-dport"].dup_rule()
        self.assertEquals(ufw.common.UFWRule.fuzzy_dst_match(x, y), 0)
        y.set_protocol("any")
        self.assertEquals(ufw.common.UFWRule.fuzzy_dst_match(x, y), -1)
        self.assertEquals(ufw.common.UFWRule.fuzzy_dst_match(y, x), 1)

        x = self.rules["multi-dport"].dup_rule()
        y = self.rules["multi-dport"].dup_rule()
        self.assertEquals(ufw.common.UFWRule.fuzzy_dst_match(x, y), 0)
        y.set_protocol("any")
        y.set_port("%s,8181" % y.dport, "dst")
        self.assertEquals(ufw.common.UFWRule.fuzzy_dst_match(x, y), 1)
        self.assertEquals(ufw.common.UFWRule.fuzzy_dst_match(y, x), 1)

        x = self.rules["any"].dup_rule()
        x.set_port("80")
        x.set_protocol("tcp")
        y = self.rules["multi-dport"].dup_rule()
        self.assertEquals(ufw.common.UFWRule.fuzzy_dst_match(x, y), -1)
        self.assertEquals(ufw.common.UFWRule.fuzzy_dst_match(y, x), 1)

        x = self.rules["any"].dup_rule()
        x.set_port("8081")
        x.set_protocol("tcp")
        y = self.rules["multi-dport"].dup_rule()
        self.assertEquals(ufw.common.UFWRule.fuzzy_dst_match(x, y), -1)
        self.assertEquals(ufw.common.UFWRule.fuzzy_dst_match(y, x), 1)

        x = self.rules["any"].dup_rule()
        x.set_port("8079")
        x.set_protocol("tcp")
        y = self.rules["multi-dport"].dup_rule()
        self.assertEquals(ufw.common.UFWRule.fuzzy_dst_match(x, y), 1)
        self.assertEquals(ufw.common.UFWRule.fuzzy_dst_match(y, x), 1)

        x = self.rules["full-any"].dup_rule()
        y = self.rules["full-any"].dup_rule()
        self.assertEquals(ufw.common.UFWRule.fuzzy_dst_match(x, y), 0)
        y.set_direction("out")
        self.assertEquals(ufw.common.UFWRule.fuzzy_dst_match(x, y), 1)

        x = self.rules["full-any"].dup_rule()
        y = self.rules["full-any"].dup_rule()
        y.set_dst("10.0.0.3")
        self.assertEquals(ufw.common.UFWRule.fuzzy_dst_match(x, y), 1)
        self.assertEquals(ufw.common.UFWRule.fuzzy_dst_match(y, x), 1)

        x = self.rules["full-any"].dup_rule()
        y = self.rules["full-any"].dup_rule()
        y.set_dst("11.0.0.0/8")
        self.assertEquals(ufw.common.UFWRule.fuzzy_dst_match(x, y), 1)
        self.assertEquals(ufw.common.UFWRule.fuzzy_dst_match(y, x), 1)

        x = self.rules["full-any"].dup_rule()
        y = self.rules["full-any"].dup_rule()
        y.set_interface("in", "eth0")
        self.assertEquals(ufw.common.UFWRule.fuzzy_dst_match(x, y), 1)
        self.assertEquals(ufw.common.UFWRule.fuzzy_dst_match(y, x), -1)

        x = self.rules["full-any"].dup_rule()
        x.set_interface("in", "eth0")
        y = x.dup_rule()
        self.assertEquals(ufw.common.UFWRule.fuzzy_dst_match(x, y), 0)
        self.assertEquals(ufw.common.UFWRule.fuzzy_dst_match(y, x), 0)

        x = self.rules["full-any"].dup_rule()
        x.set_interface("in", "eth0")
        y = x.dup_rule()
        y.set_interface("in", "eth1")
        self.assertEquals(ufw.common.UFWRule.fuzzy_dst_match(x, y), 1)
        self.assertEquals(ufw.common.UFWRule.fuzzy_dst_match(y, x), 1)

        x = self.rules["full-any"].dup_rule()
        x.set_interface("in", "lo")
        y = x.dup_rule()
        self.assertEquals(ufw.common.UFWRule.fuzzy_dst_match(x, y), 0)
        self.assertEquals(ufw.common.UFWRule.fuzzy_dst_match(y, x), 0)

        x = self.rules["full-any"].dup_rule()
        x.set_interface("in", "lo")
        y = x.dup_rule()
        y.set_dst("11.0.0.0/8")
        self.assertEquals(ufw.common.UFWRule.fuzzy_dst_match(x, y), 1)
        self.assertEquals(ufw.common.UFWRule.fuzzy_dst_match(y, x), 1)

        x = self.rules["any"].dup_rule()
        y = x.dup_rule()
        y.set_v6(True)
        self.assertEquals(ufw.common.UFWRule.fuzzy_dst_match(x, y), 1)
        self.assertEquals(ufw.common.UFWRule.fuzzy_dst_match(y, x), 1)

        x = self.rules["any"].dup_rule()
        y = x.dup_rule()
        y.forward = True
        self.assertEquals(ufw.common.UFWRule.fuzzy_dst_match(x, y), 1)
        self.assertEquals(ufw.common.UFWRule.fuzzy_dst_match(y, x), 1)

        tests.unit.support.check_for_exception(self, ValueError,
                                               x.fuzzy_dst_match,
                                               None)
    def test__is_anywhere(self):
        '''Test _is_anywhere()'''
        r = self.rules['any']
        self.assertTrue(r._is_anywhere("::/0"))
        self.assertTrue(r._is_anywhere("0.0.0.0/0"))
        self.assertFalse(r._is_anywhere("::1"))
        self.assertFalse(r._is_anywhere("fe80::1/16"))
        self.assertFalse(r._is_anywhere("127.0.0.1"))
        self.assertFalse(r._is_anywhere("127.0.0.1/32"))

    def test_get_app_tuple(self):
        '''Test get_app_tuple()'''
        r = self.rules['dapp'].dup_rule()
        t = r.get_app_tuple().split()
        self.assertEquals(self.rules['dapp'].dapp, t[0])
        self.assertEquals(self.rules['dapp'].dst, t[1])
        self.assertEquals("any", t[2])
        self.assertEquals("0.0.0.0/0", t[3])
        r.set_interface("in", "eth0")
        t = r.get_app_tuple().split()
        self.assertEquals(self.rules['dapp'].dapp, t[0])
        self.assertEquals(self.rules['dapp'].dst, t[1])
        self.assertEquals("any", t[2])
        self.assertEquals("0.0.0.0/0", t[3])
        self.assertEquals("in_eth0", t[4])

        r = self.rules['sapp'].dup_rule()
        t = r.get_app_tuple().split()
        self.assertEquals("any", t[0])
        self.assertEquals("0.0.0.0/0", t[1])
        self.assertEquals(self.rules['sapp'].sapp, t[2])
        self.assertEquals(self.rules['sapp'].src, t[3])
        r.set_interface("out", "eth0")
        t = r.get_app_tuple().split()
        self.assertEquals("any", t[0])
        self.assertEquals("0.0.0.0/0", t[1])
        self.assertEquals(self.rules['sapp'].sapp, t[2])
        self.assertEquals(self.rules['sapp'].src, t[3])
        self.assertEquals("out_eth0", t[4])

        # also test with '_' in the name (LP: #1098472)
        r = self.rules['sapp'].dup_rule()
        t = r.get_app_tuple().split()
        self.assertEquals("any", t[0])
        self.assertEquals("0.0.0.0/0", t[1])
        self.assertEquals(self.rules['sapp'].sapp, t[2])
        self.assertEquals(self.rules['sapp'].src, t[3])
        r.set_interface("out", "br_lan")
        t = r.get_app_tuple().split()
        self.assertEquals("any", t[0])
        self.assertEquals("0.0.0.0/0", t[1])
        self.assertEquals(self.rules['sapp'].sapp, t[2])
        self.assertEquals(self.rules['sapp'].src, t[3])
        self.assertEquals("out_br_lan", t[4])

def test_main(): # used by runner.py
    tests.unit.support.run_unittest(
            CommonTestCase
    )

if __name__ == "__main__": # used when standalone
    unittest.main()



'''util.py: utility functions for ufw'''
#
# Copyright 2008-2013 Canonical Ltd.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License version 3,
#    as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import print_function
import binascii
import codecs
import errno
import fcntl
import io
import inspect
import os
import re
import shutil
import socket
import struct
import subprocess
import sys

from functools import reduce
from tempfile import mkstemp

DEBUGGING = False
msg_output = None # for redirecting stdout in msg() and write_to_file()

# We support different protocols these days and only come combinations are
# valid
supported_protocols = [ 'tcp', 'udp', 'ipv6', 'esp', 'ah', 'igmp', 'gre' ]
portless_protocols = [ 'ipv6', 'esp', 'ah', 'igmp', 'gre' ]
ipv4_only_protocols = [ 'ipv6', 'igmp' ]

def get_services_proto(port):
    '''Get the protocol for a specified port from /etc/services'''
    proto = ""
    try:
        socket.getservbyname(port)
    except Exception:
        raise

    try:
        socket.getservbyname(port, "tcp")
        proto = "tcp"
    except Exception:
        pass

    try:
        socket.getservbyname(port, "udp")
        if proto == "tcp":
            proto = "any"
        else:
            proto = "udp"
    except Exception:
        pass

    return proto


def parse_port_proto(p_str):
    '''Parse port or port and protocol'''
    port = ""
    proto = ""
    tmp = p_str.split('/')
    if len(tmp) == 1:
        port = tmp[0]
        proto = "any"
    elif len(tmp) == 2:
        port = tmp[0]
        proto = tmp[1]
        if proto in portless_protocols:
            err_msg = _("Invalid port with protocol '%s'" % proto)
            raise ValueError(err_msg)
    else:
        err_msg = _("Bad port")
        raise ValueError(err_msg)
    return (port, proto)


def valid_address6(addr):
    '''Verifies if valid IPv6 address'''
    if not socket.has_ipv6:
        warn("python does not have IPv6 support.")
        return False

    # quick and dirty test
    if len(addr) > 43 or not re.match(r'^[a-fA-F0-9:\./]+$', addr):
        return False

    net = addr.split('/')
    try:
        socket.inet_pton(socket.AF_INET6, net[0])
    except Exception:
        return False

    if len(net) > 2:
        return False
    elif len(net) == 2:
        # Check netmask specified via '/'
        if not _valid_cidr_netmask(net[1], True):
            return False

    return True


def valid_address4(addr):
    '''Verifies if valid IPv4 address'''
    # quick and dirty test
    if len(addr) > 31 or not re.match(r'^[0-9\./]+$', addr):
        return False

    net = addr.split('/')
    try:
        socket.inet_pton(socket.AF_INET, net[0])
        # socket.inet_pton() should raise an exception, but let's be sure
        if not _valid_dotted_quads(net[0], False): # pragma: no cover
            return False
    except Exception:
        return False

    if len(net) > 2:
        return False
    elif len(net) == 2:
        # Check netmask specified via '/'
        if not valid_netmask(net[1], False):
            return False

    return True


def valid_netmask(nm, v6):
    '''Verifies if valid cidr or dotted netmask'''
    return _valid_cidr_netmask(nm, v6) or _valid_dotted_quads(nm, v6)


#
# valid_address()
#    version="6" tests if a valid IPv6 address
#    version="4" tests if a valid IPv4 address
#    version="any" tests if a valid IP address (IPv4 or IPv6)
#
def valid_address(addr, version="any"):
    '''Validate IP addresses'''
    if version == "6":
        return valid_address6(addr)
    elif version == "4":
        return valid_address4(addr)
    elif version == "any":
        return valid_address4(addr) or valid_address6(addr)

    raise ValueError


def normalize_address(orig, v6):
    '''Convert address to standard form. Use no netmask for IP addresses. If
       netmask is specified and not all 1's, for IPv4 use cidr if possible,
       otherwise dotted netmask and for IPv6, use cidr.
    '''
    net = []
    changed = False
    version = "4"
    s_type = socket.AF_INET
    if v6:
        version = "6"
        s_type = socket.AF_INET6

    if '/' in orig:
        net = orig.split('/')
        # Remove host netmasks
        if v6 and net[1] == "128":
            del net[1]
        elif not v6 and (net[1] == "32" or net[1] == "255.255.255.255"):
            del net[1]
    else:
        net.append(orig)

    if not v6 and len(net) == 2 and _valid_dotted_quads(net[1], v6):
        try:
            net[1] = _dotted_netmask_to_cidr(net[1], v6)
        except Exception:
            # Not valid cidr, so just use the dotted quads
            pass

    addr = net[0]

    # Convert to packed binary, then convert back
    addr = socket.inet_ntop(s_type, socket.inet_pton(s_type, addr))
    if addr != net[0]:
        changed = True

    if len(net) == 2:
        addr += "/" + net[1]
        if not v6:
            network = _address4_to_network(addr)
            if network != addr:
                dbg_msg = "Using '%s' for address '%s'" % (network, addr)
                debug(dbg_msg)
                addr = network
                changed = True

    if not valid_address(addr, version):
        dbg_msg = "Invalid address '%s'" % (addr)
        debug(dbg_msg)
        raise ValueError

    return (addr, changed)


def open_file_read(fn):
    '''Opens the specified file read-only'''
    try:
        orig = open(fn, 'r')
    except Exception:
        raise

    return orig


def open_files(fn):
    '''Opens the specified file read-only and a tempfile read-write.'''
    try:
        orig = open_file_read(fn)
    except Exception:
        raise

    try:
        (tmp, tmpname) = mkstemp()
    except Exception: # pragma: no cover
        orig.close()
        raise

    return { "orig": orig, "origname": fn, "tmp": tmp, "tmpname": tmpname }


def write_to_file(fd, out):
    '''Write to the file descriptor and error out of 0 bytes written. Intended
       to be used with open_files() and close_files().'''
    if out == "":
        return

    if not fd:
        raise OSError(errno.ENOENT, "Not a valid file descriptor")

    # Redirect our writes to stdout to msg_output, if it is set
    if msg_output and fd == sys.stdout.fileno():
        msg_output.write(out)
        return

    rc = -1
    # cover not in python3, so can't test for this
    if sys.version_info[0] >= 3: # pragma: no cover
        rc = os.write(fd, bytes(out, 'ascii'))
    else:
        rc = os.write(fd, out)

    if rc <= 0: # pragma: no cover
        raise OSError(errno.EIO, "Could not write to file descriptor")


def close_files(fns, update = True):
    '''Closes the specified files (as returned by open_files), and update
       original file with the temporary file.
    '''
    fns['orig'].close()
    os.close(fns['tmp'])

    if update:
        try:
            shutil.copystat(fns['origname'], fns['tmpname'])
            shutil.copy(fns['tmpname'], fns['origname'])
        except Exception:
            raise

    try:
        os.unlink(fns['tmpname'])
    except OSError:
        raise


def cmd(command):
    '''Try to execute the given command.'''
    debug(command)
    try:
        sp = subprocess.Popen(command, stdout=subprocess.PIPE,
                              stderr=subprocess.STDOUT,
                              universal_newlines=True)
    except OSError as ex:
        return [127, str(ex)]

    out = sp.communicate()[0]
    return [sp.returncode, str(out)]


def cmd_pipe(command1, command2):
    '''Try to pipe command1 into command2.'''
    try:
        sp1 = subprocess.Popen(command1, stdout=subprocess.PIPE)
        sp2 = subprocess.Popen(command2, stdin=sp1.stdout)
    except OSError as ex:
        return [127, str(ex)]

    out = sp2.communicate()[0]
    return [sp2.returncode, str(out)]

# TODO: this is pretty horrible. We should be using only unicode strings
#       internally and decode() when printing rather than doing this.
def _print(output, s):
    '''Implement our own print statement that will output utf-8 when
       appropriate.'''
    try: # python3
        writer = output.buffer
    except:
        writer = output

    try:
        out = s.encode('utf-8', 'ignore')
    # Depends on python version
    except: # pragma: no cover
        out = s

    if msg_output and inspect.isclass(io.StringIO):
        writer.write(s)
    else:
        writer.write(bytes(out))
    output.flush()


def error(out, do_exit=True):
    '''Print error message and exit'''
    try:
        _print(sys.stderr, 'ERROR: %s\n' % out)
    except IOError: # pragma: no cover
        pass

    if do_exit: # pragma: no cover
        sys.exit(1)


def warn(out):
    '''Print warning message'''
    try:
        _print(sys.stderr, 'WARN: %s\n' % out)
    except IOError: # pragma: no cover
        pass


def msg(out, output=sys.stdout, newline=True):
    '''Print message'''
    if msg_output and output == sys.stdout:
        output = msg_output

    try:
        if newline:
            _print(output, '%s\n' % out)
        else:
            _print(output, '%s' % out)
    except IOError: # pragma: no cover
        pass


def debug(out):
    '''Print debug message'''
    if DEBUGGING:
        try:
            _print(sys.stderr, 'DEBUG: %s\n' % out)
        except IOError: # pragma: no cover
            pass


def word_wrap(text, width):
    '''
    A word-wrap function that preserves existing line breaks
    and most spaces in the text. Expects that existing line
    breaks are posix newlines (\n).
    '''
    return reduce(lambda line, word, width=width: '%s%s%s' %
                  (line,
                   ' \n'[(len(line)-line.rfind('\n')-1
                         + len(word.split('\n',1)[0]
                              ) >= width)],
                   word),
                  text.split(' ')
                 )


def wrap_text(text):
    '''Word wrap to a specific width'''
    return word_wrap(text, 75)


def human_sort(lst):
    '''Sorts list of strings into numeric order, with text case-insensitive.
       Modifies list in place.

       Eg:
       [ '80', 'a222', 'a32', 'a2', 'b1', '443', 'telnet', '3', 'http', 'ZZZ']

       sorts to:
       ['3', '80', '443', 'a2', 'a32', 'a222', 'b1', 'http', 'telnet', 'ZZZ']
    '''
    norm = lambda t: int(t) if t.isdigit() else t.lower()
    lst.sort(key=lambda k: [ norm(c) for c in re.split('([0-9]+)', k)])


def get_ppid(mypid=os.getpid()):
    '''Finds parent process id for pid based on /proc/<pid>/stat. See
       'man 5 proc' for details.
    '''
    try:
        pid = int(mypid)
    except Exception:
        raise ValueError("pid must be an integer")

    name = os.path.join("/proc", str(pid), "stat")
    if not os.path.isfile(name):
        raise IOError("Couldn't find '%s'" % (name))

    try:
        # LP: #1101304
        # 9983 (cmd) S 923 ...
        # 9983 (cmd with spaces) S 923 ...
        ppid = open(name).readlines()[0].split(')')[1].split()[1]
    except Exception: # pragma: no cover
        raise

    return int(ppid)


def under_ssh(pid=os.getpid()):
    '''Determine if current process is running under ssh'''
    try:
        ppid = get_ppid(pid)
    except IOError:
        warn_msg = _("Couldn't find pid (is /proc mounted?)")
        warn(warn_msg)
        return False
    except Exception:
        err_msg = _("Couldn't find parent pid for '%s'") % (str(pid))
        raise ValueError(err_msg)

    # pid '1' is 'init' and '0' is the kernel. This should still work when
    # pid randomization is in use, but needs to be checked.
    if pid == 1 or ppid <= 1:
        return False

    path = os.path.join("/proc", str(ppid), "stat")
    if not os.path.isfile(path): # pragma: no cover
        err_msg = _("Couldn't find '%s'") % (path)
        raise ValueError(err_msg)

    try:
        exe = open(path).readlines()[0].split()[1]
    except Exception: # pragma: no cover
        err_msg = _("Could not find executable for '%s'") % (path)
        raise ValueError(err_msg)
    debug("under_ssh: exe is '%s'" % (exe))

    # unit tests might be run remotely, so can't test for either
    if exe == "(sshd)": # pragma: no cover
        return True
    else: # pragma: no cover
        return under_ssh(ppid)


#
# Internal helper functions
#
def _valid_cidr_netmask(nm, v6):
    '''Verifies cidr netmasks'''
    num = 32
    if v6:
        num = 128

    if not re.match(r'^[0-9]+$', nm) or int(nm) < 0 or int(nm) > num:
        return False

    return True


def _valid_dotted_quads(nm, v6):
    '''Verifies dotted quad ip addresses and netmasks'''
    if v6:
        return False
    else:
        if re.match(r'^[0-9]+\.[0-9\.]+$', nm):
            quads = re.split('\.', nm)
            if len(quads) != 4:
                return False
            for q in quads:
                if not q or int(q) < 0 or int(q) > 255:
                    return False
        else:
            return False

    return True


#
# _dotted_netmask_to_cidr()
# Returns:
#   cidr integer (0-32 for ipv4 and 0-128 for ipv6)
#
# Raises exception if cidr cannot be found
#
def _dotted_netmask_to_cidr(nm, v6):
    '''Convert netmask to cidr. IPv6 dotted netmasks are not supported.'''
    cidr = ""
    if v6:
        raise ValueError
    else:
        if not _valid_dotted_quads(nm, v6):
            raise ValueError

        mbits = 0

        # python3 doesn't have long(). We could technically use int() here
        # since python2 guarantees at least 32 bits for int(), but this helps
        # future-proof.
        try: # pragma: no cover
            bits = long(struct.unpack('>L', socket.inet_aton(nm))[0])
        except NameError: # pragma: no cover
            bits = int(struct.unpack('>L', socket.inet_aton(nm))[0])

        found_one = False
        for n in range(32):
            if (bits >> n) & 1 == 1:
                found_one = True
            else:
                if found_one:
                    mbits = -1
                    break
                else:
                    mbits += 1

        if mbits >= 0 and mbits <= 32:
            cidr = str(32 - mbits)

    if not _valid_cidr_netmask(cidr, v6):
        raise ValueError

    return cidr


#
# _cidr_to_dotted_netmask()
# Returns:
#   dotted netmask string
#
# Raises exception if dotted netmask cannot be found
#
def _cidr_to_dotted_netmask(cidr, v6):
    '''Convert cidr to netmask. IPv6 dotted netmasks not supported.'''
    nm = ""
    if v6:
        raise ValueError
    else:
        if not _valid_cidr_netmask(cidr, v6):
            raise ValueError

        # python3 doesn't have long(). We could technically use int() here
        # since python2 guarantees at least 32 bits for int(), but this helps
        # future-proof.
        try: # pragma: no cover
            bits = long(0)
        except NameError: # pragma: no cover
            bits = 0

        for n in range(32):
            if n < int(cidr):
                bits |= 1 << 31 - n
        nm = socket.inet_ntoa(struct.pack('>L', bits))

    # The above socket.inet_ntoa() should raise an error, but let's be sure
    if not _valid_dotted_quads(nm, v6): # pragma: no cover
        raise ValueError

    return nm


def _address4_to_network(addr):
    '''Convert an IPv4 address and netmask to a network address'''
    if '/' not in addr:
        debug("_address4_to_network: skipping address without a netmask")
        return addr

    tmp = addr.split('/')
    if len(tmp) != 2 or not _valid_dotted_quads(tmp[0], False):
        raise ValueError

    host = tmp[0]
    orig_nm = tmp[1]

    nm = orig_nm
    if _valid_cidr_netmask(nm, False):
        nm = _cidr_to_dotted_netmask(nm, False)

    # Now have dotted quad host and nm, find the network

    # python3 doesn't have long(). We could technically use int() here
    # since python2 guarantees at least 32 bits for int(), but this helps
    # future-proof.
    try: # pragma: no cover
        host_bits = long(struct.unpack('>L', socket.inet_aton(host))[0])
        nm_bits = long(struct.unpack('>L', socket.inet_aton(nm))[0])
    except NameError: # pragma: no cover
        host_bits = int(struct.unpack('>L', socket.inet_aton(host))[0])
        nm_bits = int(struct.unpack('>L', socket.inet_aton(nm))[0])

    network_bits = host_bits & nm_bits
    network = socket.inet_ntoa(struct.pack('>L', network_bits))

    return "%s/%s" % (network, orig_nm)

def _address6_to_network(addr):
    '''Convert an IPv6 address and netmask to a network address'''
    def dec2bin(num, count):
        '''Decimal to binary'''
        return "".join([str((num >> y) & 1) for y in range(count-1, -1, -1)])

    if '/' not in addr:
        debug("_address6_to_network: skipping address without a netmask")
        return addr

    tmp = addr.split('/')
    if len(tmp) != 2 or not valid_netmask(tmp[1], True):
        raise ValueError

    orig_host = tmp[0]
    netmask = tmp[1]

    unpacked = struct.unpack('>8H', socket.inet_pton(socket.AF_INET6, \
                                                     orig_host))

    # Get the host bits
    try: # python3 doesn't have long()
        host_bits = long(0)
    except NameError: # pragma: no cover
        host_bits = 0

    for i in range(8):
        n = dec2bin(unpacked[i], 16)
        for j in range(16):
            host_bits |= (1 & int(n[j])) <<(127-j-i*16)

    # Create netmask bits
    try: # python3 doesn't have long()
        nm_bits = long(0)
    except NameError: # pragma: no cover
        nm_bits = 0

    for i in range(128):
        if i < int(netmask):
            nm_bits |= 1 << (128 - 1) - i

    # Apply the netmask to the host to determine the network
    net = host_bits & nm_bits

    # Break the network into chunks suitable for repacking
    lst = []
    for i in range(8):
        lst.append(int(dec2bin(net, 128)[i*16:i*16+16], 2))

    # Create the network string
    network = socket.inet_ntop(socket.AF_INET6, \
                               struct.pack('>8H', lst[0], lst[1], \
                                           lst[2], lst[3], lst[4], \
                                           lst[5], lst[6], lst[7]))

    return "%s/%s" % (network, netmask)


def in_network(tested_add, tested_net, v6):
    '''Determine if address x is in network y'''
    tmp = tested_net.split('/')
    if len(tmp) != 2 or not valid_netmask(tmp[1], v6):
        raise ValueError

    orig_host = tmp[0]
    netmask = tmp[1]

    if orig_host == "0.0.0.0" or orig_host == "::":
        return True

    address = tested_add
    if '/' in address:
        tmp = address.split('/')
        if len(tmp) != 2 or not valid_netmask(tmp[1], v6):
            raise ValueError
        address = tmp[0]

    if address == "0.0.0.0" or address == "::":
        return True

    if v6:
        if not valid_address6(address) or not valid_address6(orig_host):
            raise ValueError
    else:
        if not valid_address4(address) or not valid_address4(orig_host):
            raise ValueError

    if _valid_cidr_netmask(netmask, v6) and not v6:
        netmask = _cidr_to_dotted_netmask(netmask, v6)

    # Now apply the network's netmask to the address
    if v6:
        orig_network = _address6_to_network("%s/%s" % \
                                            (orig_host, netmask)).split('/')[0]
        network = _address6_to_network("%s/%s" % \
                                       (address, netmask)).split('/')[0]
    else:
        orig_network = _address4_to_network("%s/%s" % \
                                            (orig_host, netmask)).split('/')[0]
        network = _address4_to_network("%s/%s" % \
                                       (address, netmask)).split('/')[0]

    return network == orig_network


def get_iptables_version(exe="/sbin/iptables"):
    '''Return iptables version'''
    (rc, out) = cmd([exe, '-V'])
    if rc != 0:
        raise OSError(errno.ENOENT, "Error running '%s'" % (exe))
    tmp = re.split('\s', out)
    return re.sub('^v', '', tmp[1])


# must be root, so don't report coverage in unit tests
def get_netfilter_capabilities(exe="/sbin/iptables", do_checks=True):
    '''Return capabilities set for netfilter to support new features. Callers
       must be root.'''
    def test_cap(exe, chain, rule):
        args = [exe, '-A', chain]
        (rc, out) = cmd(args + rule)
        if rc == 0:
            return True
        return False # pragma: no cover

    if do_checks and os.getuid() != 0:
        raise OSError(errno.EPERM, "Must be root")

    caps = []

    chain = "ufw-caps-test"
    if exe.endswith("ip6tables"):
        chain = "ufw6-caps-test"

    # First install a test chain
    (rc, out) = cmd([exe, '-N', chain])
    if rc != 0:
        raise OSError(errno.ENOENT, out) # pragma: no cover

    # Now test for various capabilities. We won't test for everything, just
    # the stuff we know isn't supported everywhere but we want to support.

    # recent-set
    if test_cap(exe, chain, ['-m', 'conntrack', '--ctstate', 'NEW', \
                             '-m', 'recent', '--set']):
        caps.append('recent-set')

    # recent-update
    if test_cap(exe, chain, ['-m', 'conntrack', '--ctstate', 'NEW', \
                             '-m', 'recent', '--update', \
                             '--seconds', '30', \
                             '--hitcount', '6']):
        caps.append('recent-update')

    # Cleanup
    cmd([exe, '-F', chain])
    (rc, out) = cmd([exe, '-X', chain])
    if rc != 0:
        raise OSError(errno.ENOENT, out) # pragma: no cover

    return caps

def parse_netstat_output(v6):
    '''Get and parse netstat the output from get_netstat_output()'''

    # d[proto][port] -> list of dicts:
    #   d[proto][port][0][laddr|raddr|uid|pid|exe]

    netstat_output = get_netstat_output(v6)

    d = dict()
    for line in netstat_output.splitlines():
        if not line.startswith('tcp') and not line.startswith('udp'): # pragma: no cover
            continue

        tmp = line.split()

        proto = tmp[0]
        port  = tmp[1].split(':')[-1]

        item = dict()
        item['laddr'] = ':'.join(tmp[1].split(':')[:-1])
        item['uid']   = tmp[3]
        item['pid'] = tmp[5].split('/')[0]
        if item['pid'] == '-':
            item['exe'] = item['pid']
        else: # pragma: no cover
            item['exe'] = tmp[5].split('/')[1]

        if proto not in d:
            d[proto] = dict()
            d[proto][port] = []
        else:
            if port not in d[proto]:
                d[proto][port] = []
        d[proto][port].append(item)

    return d


def get_ip_from_if(ifname, v6=False):
    '''Get IP address for interface'''
    addr = ""

    # we may not have an IPv6 address, so no coverage
    if v6: # pragma: no cover
        proc = '/proc/net/if_inet6'
        if not os.path.exists(proc):
            raise OSError(errno.ENOENT, "'%s' does not exist" % proc)

        for line in open(proc).readlines():
            tmp = line.split()
            if ifname == tmp[5]:
                addr = ":".join( \
                           [ tmp[0][i:i+4] for i in range(0,len(tmp[0]),4) ])

                if tmp[2].lower() != "80":
                    addr = "%s/%s" % (addr, int(tmp[2].lower(), 16))

        if addr == "":
            raise IOError(errno.ENODEV, "No such device")
    else:
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            addr = socket.inet_ntoa(fcntl.ioctl(s.fileno(), 0x8915, \
                                    struct.pack('256s', ifname[:15]))[20:24])
        except Exception:
            raise IOError(errno.ENODEV, "No such device")

    return normalize_address(addr, v6)[0]


def get_if_from_ip(addr):
    '''Get interface for IP address'''
    v6 = False
    proc = '/proc/net/dev'
    if valid_address6(addr):
        v6 = True
        proc = '/proc/net/if_inet6'
    elif not valid_address4(addr):
        raise IOError(errno.ENODEV, "No such device")

    if not os.path.exists(proc): # pragma: no cover
        raise OSError(errno.ENOENT, "'%s' does not exist" % proc)

    matched = ""
    # we may not have an IPv6 address, so no coverage
    if v6: # pragma: no cover
        for line in open(proc).readlines():
            tmp = line.split()
            ifname = tmp[5].strip()

            tmp_addr = ":".join( \
                           [ tmp[0][i:i+4] for i in range(0,len(tmp[0]),4) ])
            if tmp[2].lower() != "80":
                tmp_addr = "%s/%s" % (tmp_addr, int(tmp[2].lower(), 16))

            if addr == tmp_addr or \
               ('/' in tmp_addr and in_network(addr, tmp_addr, True)):
                matched = ifname
                break
    else:
        for line in open(proc).readlines():
            if ':' not in line:
                continue
            ifname = line.split(':')[0].strip()
            # this can fail for certain devices, so just skip them
            try:
                ip = get_ip_from_if(ifname, False)
            except IOError: # pragma: no cover
                continue

            if ip == addr:
                matched = ifname
                break

    return matched


def _get_proc_inodes():
    '''Get inodes of files in /proc'''
    proc_files = os.listdir("/proc")
    proc_files.sort()
    pat = re.compile(r'^[0-9]+$')
    inodes = dict()
    for i in proc_files:
        if not pat.match(i):
            continue

        fd_path = os.path.join("/proc", i, "fd")

        # skip stuff we can't read or that goes away
        if not os.access(fd_path, os.F_OK | os.R_OK):
            continue

        exe_path = "-"
        try:
            exe_path = os.readlink(os.path.join("/proc", i, "exe"))
        except Exception: # pragma: no cover
            pass

        try:
            dirs = os.listdir(fd_path)
        except: # pragma: no cover
            continue

        for j in dirs:
            try:
                inode = os.stat(os.path.join(fd_path, j))[1]
            except Exception: # pragma: no cover
                continue
            inodes[inode] = "%s/%s" % (i, os.path.basename(exe_path))

    return inodes


def _read_proc_net_protocol(protocol):
    '''Read /proc/net/(tcp|udp)[6] file and return a list of tuples '''
    tcp_states = { 1: "ESTABLISHED",
                   2: "SYN_SENT",
                   3: "SYN_RECV",
                   4: "FIN_WAIT1",
                   5: "FIN_WAIT2",
                   6: "TIME_WAIT",
                   7: "CLOSE",
                   8: "CLOSE_WAIT",
                   9: "LAST_ACK",
                   10: "LISTEN",
                   11: "CLOSING"
                 }

    proc_net_fields = { 'local_addr': 1,
                        'state': 3,
                        'uid': 7,
                        'inode': 9
                      }

    fn = os.path.join("/proc/net", protocol)
    # can't test for this
    if not os.access(fn, os.F_OK | os.R_OK): # pragma: no cover
        raise ValueError

    lst = []
    skipped_first = False
    lines = open(fn).readlines()
    for line in lines:
        fields = line.split()
        if not skipped_first:
            skipped_first = True
            continue
        state = tcp_states[int(fields[proc_net_fields['state']], 16)]
        if protocol.startswith("udp"):
            state = "NA"
        elif protocol.startswith("tcp") and state != "LISTEN":
            continue
        laddr, port = fields[proc_net_fields['local_addr']].split(':')
        uid = fields[proc_net_fields['uid']]
        inode = fields[proc_net_fields['inode']]
        lst.append((laddr, int(port, 16), uid, inode, state))

    return lst


def convert_proc_address(paddr):
    '''Convert an address from /proc/net/(tcp|udp)* to a normalized address'''
    converted = ""
    if len(paddr) > 8:
        tmp = ""
        for i in range(0, 32, 8):
            tmp += "".join([ paddr[j-2:j] for j in range(i+8, i, -2) ])
        converted = normalize_address(":".join( \
               [ tmp[j:j+4].lower() for j in range(0,len(tmp),4) ]), \
               True)[0]
    else:
        tmp = []
        for i in [ paddr[j-2:j] for j in range(8, 0, -2) ]:
            tmp.append(str(int(i, 16)))
        converted = normalize_address(".".join(tmp), False)[0]

    return converted


def get_netstat_output(v6):
    '''netstat-style output, without IPv6 address truncation'''
    proc_net_data = dict()
    proto = ['tcp', 'udp']
    if v6:
        proto += ['tcp6', 'udp6']
    for p in proto:
        try:
            proc_net_data[p] = _read_proc_net_protocol(p)
        except Exception: # pragma: no cover
            warn_msg = _("Could not get statistics for '%s'" % (p))
            warn(warn_msg)
            continue

    inodes = _get_proc_inodes()

    protocols = list(proc_net_data.keys())
    protocols.sort()

    s = ""
    for p in protocols:
        for (laddr, port, uid, inode, state) in proc_net_data[p]:
            addr = convert_proc_address(laddr)

            exe = "-"
            if int(inode) in inodes:
                # need root for this, so turn off in unit tests
                exe = inodes[int(inode)] # pragma: no cover
            s += "%-5s %-46s %-11s %-5s %-11s %s\n" % (p,
                                                       "%s:%s" % (addr, port),
                                                       state, uid, inode, exe)

    return s

def _findpath(dir, prefix):
    '''Add prefix to dir'''
    if prefix is None:
        return dir
    if dir.startswith('/'):
        if len(dir) < 2:  # /
            newdir = prefix
        else:
            newdir = os.path.join(prefix, dir[1:])
    else:
        newdir = os.path.join(prefix, dir)
    return newdir

def hex_encode(s):
    '''Take a string and convert it to a hex string'''
    if sys.version_info[0] < 3:
        return codecs.encode(s, 'hex')
    # hexlify returns a bytes string (eg, b'ab12cd') so decode that to ascii
    # to have identical output as python2
    return binascii.hexlify(s.encode('utf-8', errors='ignore')).decode('ascii')

def hex_decode(h):
    '''Take a hex string and convert it to a string'''
    if sys.version_info[0] < 3:
        return h.decode(encoding='hex').decode('utf-8')
    return binascii.unhexlify(h).decode('utf-8')
